import logging
import time

from django.core.management.base import BaseCommand
from web3 import Web3

from hub20.apps.blockchain.models import make_web3
from hub20.apps.ethereum_money.models import EthereumToken
from hub20.apps.raiden import models
from hub20.apps.raiden.client import RaidenClient, RaidenConnectionError
from hub20.apps.raiden.contracts import get_token_network_registry_contract

logger = logging.getLogger(__name__)


def sync_token_networks(client: RaidenClient, w3: Web3):
    logger.info("Updating Token Networks")
    chain_id = int(w3.net.version)
    known_tokens = client.raiden.token_networks.values_list("token__address", flat=True)

    for token_address in client.get_token_addresses():
        if token_address in known_tokens:
            continue

        logger.info(f"Getting information about token on {token_address}")
        token = EthereumToken.make(token_address, chain_id)
        token_network_registry_contract = get_token_network_registry_contract(w3)
        token_network = models.TokenNetwork.make(token, token_network_registry_contract)
        client.raiden.token_networks.add(token_network)


def sync_channels(client: RaidenClient):
    logger.info("Updating Channels")
    for channel_data in client.get_channels():
        channel = models.Channel.make(client.raiden, **channel_data)
        logger.info(f"{channel} information synced")


def sync_payments(client: RaidenClient):
    for channel in client.raiden.channels.all():
        logger.info(f"Getting new payments from {channel}")
        for payment_data in client.get_new_payments(channel):
            models.Payment.make(channel, **payment_data)


class Command(BaseCommand):
    help = "Connects to Raiden via REST API to collect information about new transfers"

    def handle(self, *args, **options):
        w3 = make_web3()

        while True:
            for raiden in models.Raiden.objects.all():
                client = RaidenClient(raiden)
                try:
                    sync_token_networks(client, w3)
                    sync_channels(client)
                    sync_payments(client)
                except RaidenConnectionError as exc:
                    logger.warn(str(exc))
                    time.sleep(5)
                except Exception as exc:
                    logger.exception(exc)
            time.sleep(3)



from typing import Any, Dict, Optional

from attributedict.collections import AttributeDict
from django.conf import settings


def _get_setting(setting_name: str, default: Any) -> Optional[Dict]:
    user_settings = getattr(settings, "HUB20", {})
    return user_settings.get(setting_name, default)


TRANSFER_SETTINGS = AttributeDict(
    {"minimum_confirmations": _get_setting("TRANSFER_MININUM_CONFIRMATIONS", 10)}
)

PAYMENT_SETTINGS = AttributeDict(
    {
        "minimum_confirmations": _get_setting("PAYMENT_MININUM_CONFIRMATIONS", 5),
        "payment_lifetime": _get_setting("PAYMENT_REQUEST_LIFETIME", 15 * 60),
    }
)

API_KEY_HEADER_NAME = _get_setting("API_KEY_HEADER_NAME", "X-HUB20-API-KEY")



import factory

from hub20.apps.blockchain.factories.providers import EthereumProvider
from hub20.apps.ethereum_money.factories import EthereumTokenValueModelFactory
from hub20.apps.core import models

from .base import UserFactory

factory.Faker.add_provider(EthereumProvider)


class InternalTransferFactory(EthereumTokenValueModelFactory):
    sender = factory.SubFactory(UserFactory)
    receiver = factory.SubFactory(UserFactory)

    class Meta:
        model = models.InternalTransfer


class ExternalTransferFactory(EthereumTokenValueModelFactory):
    sender = factory.SubFactory(UserFactory)
    recipient_address = factory.Faker("ethereum_address")

    class Meta:
        model = models.ExternalTransfer


__all__ = ["InternalTransferFactory", "ExternalTransferFactory"]



from decimal import Decimal
from unittest.mock import patch

import pytest
from django.core.exceptions import ValidationError
from django.test import TestCase, override_settings
from eth_utils import is_checksum_address, to_wei

from hub20.apps.blockchain.factories import BlockFactory, TransactionFactory
from hub20.apps.core.app_settings import PAYMENT_SETTINGS, TRANSFER_SETTINGS
from hub20.apps.core.choices import PAYMENT_EVENT_TYPES, TRANSFER_EVENT_TYPES
from hub20.apps.core.factories import (
    CheckoutFactory,
    Erc20TokenPaymentOrderFactory,
    Erc20TokenUserBalanceEntryFactory,
    ExternalTransferFactory,
    InternalTransferFactory,
    StoreFactory,
    UserAccountFactory,
    WalletFactory,
)
from hub20.apps.core.models import ExternalTransfer, UserAccount, Wallet
from hub20.apps.ethereum_money.factories import Erc20TokenFactory, Erc20TransferFactory, ETHFactory
from hub20.apps.ethereum_money.models import EthereumTokenAmount, encode_transfer_data
from hub20.apps.raiden.factories import ChannelFactory, PaymentEventFactory, TokenNetworkFactory

from .base import TEST_SETTINGS


def add_eth_to_wallet(wallet: Wallet, amount: EthereumTokenAmount):
    return TransactionFactory(to_address=wallet.address, value=to_wei(amount.amount, "ether"))


def add_token_to_wallet(wallet: Wallet, amount: EthereumTokenAmount):
    transaction_data = encode_transfer_data(wallet.address, amount)
    return Erc20TransferFactory(
        to_address=amount.currency.address,
        data=transaction_data,
        value=0,
        log__data=amount.as_hex,
    )


@pytest.mark.django_db(transaction=True)
@override_settings(**TEST_SETTINGS)
class BaseTestCase(TestCase):
    pass


class BlockchainPaymentTestCase(BaseTestCase):
    def setUp(self):
        self.order = Erc20TokenPaymentOrderFactory()

    def test_transaction_sets_payment_as_received(self):
        add_token_to_wallet(self.order.payment_method.wallet, self.order.as_token_amount)
        self.assertEqual(self.order.status, PAYMENT_EVENT_TYPES.received)

    def test_transaction_creates_blockchain_payment(self):
        add_token_to_wallet(self.order.payment_method.wallet, self.order.as_token_amount)
        self.assertEqual(self.order.payment_set.count(), 1)

    def test_user_balance_is_updated_on_completed_payment(self):
        add_token_to_wallet(self.order.payment_method.wallet, self.order.as_token_amount)
        BlockFactory.create_batch(
            PAYMENT_SETTINGS.minimum_confirmations, chain=self.order.currency.chain
        )

        user_account = UserAccount(self.order.user)
        balance = user_account.get_balance(self.order.currency)
        self.assertEqual(balance.amount, self.order.amount)


class CheckoutTestCase(BaseTestCase):
    def setUp(self):
        self.checkout = CheckoutFactory()
        self.checkout.store.accepted_currencies.add(self.checkout.payment_order.currency)

    def test_payment_order_user_and_store_owner_are_the_same(self):
        self.assertEqual(self.checkout.store.owner, self.checkout.payment_order.user)

    def test_checkout_currency_must_be_accepted_by_store(self):
        self.checkout.clean()

        self.checkout.store.accepted_currencies.clear()
        with self.assertRaises(ValidationError):
            self.checkout.clean()


class RaidenPaymentTestCase(BaseTestCase):
    def setUp(self):
        token_network = TokenNetworkFactory()

        self.channel = ChannelFactory(token_network=token_network)
        self.channel.raiden.token_networks.add(token_network)

        self.order = Erc20TokenPaymentOrderFactory(currency=token_network.token)

    def test_order_payment_method_includes_raiden(self):
        self.assertIsNotNone(self.order.payment_method)
        self.assertIsNotNone(self.order.payment_method.raiden)

    def test_order_payment_method_includes_identifier(self):
        self.assertIsNotNone(self.order.payment_method)
        self.assertIsNotNone(self.order.payment_method.identifier)

    def test_payment_via_raiden_sets_order_as_paid(self):
        PaymentEventFactory(
            channel=self.channel,
            amount=self.order.amount,
            identifier=self.order.payment_method.identifier,
            receiver_address=self.channel.raiden.address,
        )
        self.assertTrue(self.order.is_finalized)
        self.assertEqual(self.order.status, PAYMENT_EVENT_TYPES.confirmed)


class StoreTestCase(BaseTestCase):
    def setUp(self):
        self.store = StoreFactory()

    def test_store_rsa_keys_are_valid_pem(self):
        self.assertIsNotNone(self.store.rsa.pk)
        self.assertTrue(type(self.store.rsa.public_key_pem) is str)
        self.assertTrue(type(self.store.rsa.private_key_pem) is str)

        self.assertTrue(self.store.rsa.public_key_pem.startswith("-----BEGIN PUBLIC KEY-----"))
        self.assertTrue(
            self.store.rsa.private_key_pem.startswith("-----BEGIN RSA PRIVATE KEY-----")
        )


class TransferTestCase(BaseTestCase):
    def setUp(self):
        self.sender_account = UserAccountFactory()
        self.receiver_account = UserAccountFactory()
        self.sender = self.sender_account.user
        self.receiver = self.receiver_account.user
        self.credit = Erc20TokenUserBalanceEntryFactory(user=self.sender)


class InternalTransferTestCase(TransferTestCase):
    def test_transfers_are_finalized_as_confirmed(self):
        transfer = InternalTransferFactory(
            sender=self.sender,
            receiver=self.receiver,
            currency=self.credit.currency,
            amount=self.credit.amount,
        )

        self.assertTrue(transfer.is_finalized)
        self.assertEqual(transfer.status, TRANSFER_EVENT_TYPES.confirmed)

    def test_transfers_change_balance(self):
        transfer = InternalTransferFactory(
            sender=self.sender,
            receiver=self.receiver,
            currency=self.credit.currency,
            amount=self.credit.amount,
        )

        self.assertTrue(transfer.is_finalized)

        sender_balance = self.sender_account.get_balance(self.credit.currency)
        receiver_balance = self.receiver_account.get_balance(self.credit.currency)

        self.assertEqual(sender_balance.amount, 0)
        self.assertEqual(receiver_balance, self.credit.as_token_amount)

    def test_transfers_fail_with_low_sender_balance(self):
        transfer = InternalTransferFactory(
            sender=self.sender,
            receiver=self.receiver,
            currency=self.credit.currency,
            amount=2 * self.credit.amount,
        )

        self.assertTrue(transfer.is_finalized)
        self.assertEqual(transfer.status, TRANSFER_EVENT_TYPES.failed)


@patch("hub20.apps.core.models.transfers.Wallet.select_for_transfer")
class ExternalTransferTestCase(TransferTestCase):
    def setUp(self):
        super().setUp()
        self.ETH = ETHFactory()
        self.fee_amount = EthereumTokenAmount(amount=Decimal("0.001"), currency=self.ETH)

    def _build_transfer(self, wallet: Wallet) -> ExternalTransfer:
        transfer = ExternalTransferFactory.build(
            sender=self.sender, currency=self.credit.currency, amount=self.credit.amount,
        )

        out_tx = TransactionFactory(
            from_address=wallet.address,
            to_address=self.credit.currency.address,
            data=encode_transfer_data(transfer.recipient_address, self.credit.as_token_amount),
        )

        with patch.object(wallet.account, "send", return_value=out_tx.hash):
            transfer.save()

        return transfer

    def test_external_transfers_fail_without_funds(self, select_for_transfer):
        select_for_transfer.return_value = None
        transfer = ExternalTransferFactory(
            sender=self.sender, currency=self.credit.currency, amount=self.credit.amount,
        )

        self.assertIsNotNone(transfer.status)
        self.assertEqual(transfer.status, TRANSFER_EVENT_TYPES.failed)

    def test_transfers_can_be_executed_with_enough_balance(self, select_for_transfer):
        wallet = WalletFactory()
        add_token_to_wallet(wallet, self.credit.as_token_amount)
        add_eth_to_wallet(wallet, self.fee_amount)

        select_for_transfer.return_value = wallet

        transfer = self._build_transfer(wallet)

        self.assertIsNotNone(transfer.status)
        self.assertFalse(transfer.is_finalized)
        self.assertEqual(transfer.status, TRANSFER_EVENT_TYPES.executed)

    def test_transfers_are_confirmed_after_block_confirmation(self, select_for_transfer):
        wallet = WalletFactory()

        add_eth_to_wallet(wallet, self.fee_amount)
        add_token_to_wallet(wallet, self.credit.as_token_amount)
        select_for_transfer.return_value = wallet

        transfer = self._build_transfer(wallet)

        self.assertIsNotNone(transfer.status)
        self.assertFalse(transfer.is_finalized)
        self.assertEqual(transfer.status, TRANSFER_EVENT_TYPES.executed)

        BlockFactory.create_batch(TRANSFER_SETTINGS.minimum_confirmations)
        self.assertTrue(transfer.is_finalized)
        self.assertEqual(transfer.status, TRANSFER_EVENT_TYPES.confirmed)


class WalletTestCase(BaseTestCase):
    def setUp(self):
        self.wallet = WalletFactory()

    def test_wallet_address_is_checksummed(self):
        self.assertTrue(is_checksum_address(self.wallet.address))

    @patch("hub20.apps.core.models.accounting.get_max_fee")
    def test_wallet_balance_selection(self, patched_fee):
        ETH = ETHFactory()
        fee_amount = EthereumTokenAmount(amount=Decimal("0.001"), currency=ETH)

        patched_fee.return_value = fee_amount

        token = Erc20TokenFactory()
        token_amount = EthereumTokenAmount(amount=Decimal("10"), currency=token)

        add_eth_to_wallet(self.wallet, fee_amount)
        add_token_to_wallet(self.wallet, token_amount)

        self.assertIsNotNone(Wallet.select_for_transfer(token_amount))
        self.assertIsNone(Wallet.select_for_transfer(2 * fee_amount))


__all__ = [
    "BlockchainPaymentTestCase",
    "CheckoutTestCase",
    "RaidenPaymentTestCase",
    "StoreTestCase",
    "TransferTestCase",
    "InternalTransferTestCase",
    "ExternalTransferTestCase",
    "WalletTestCase",
]



from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import requests
from attributedict.collections import AttributeDict
from django.utils.timezone import make_aware

from . import models


def _make_request(url: str, method: str = "GET", **params: Any) -> Union[List, Dict]:
    action = {
        "GET": requests.get,
        "PUT": requests.put,
        "POST": requests.post,
        "DELETE": requests.delete,
    }[method.upper()]

    try:
        response = action(url)
        response.raise_for_status()
        return response.json()
    except requests.exceptions.ConnectionError:
        raise RaidenConnectionError(f"Could not connect to {url}")


class RaidenConnectionError(Exception):
    pass


class RaidenClient:
    def __init__(self, raiden: models.Raiden) -> None:
        self.raiden = raiden

    def _parse_payment(
        self, payment_data: Dict, channel: models.Channel
    ) -> Optional[AttributeDict]:
        event_name = payment_data.pop("event")
        payment_data.pop("token_address", None)

        if event_name == "EventPaymentReceivedSuccess":
            payment_data["sender_address"] = payment_data.pop("initiator")
            payment_data["receiver_address"] = self.raiden.address
        elif event_name == "EventPaymentSentSuccess":
            payment_data["sender_address"] = self.raiden.address
            payment_data["receiver_address"] = payment_data.pop("target")
        else:
            return None

        iso_time = payment_data.pop("log_time")

        payment_data["amount"] = channel.token.from_wei(payment_data.pop("amount")).amount
        payment_data["timestamp"] = make_aware(datetime.fromisoformat(iso_time))
        return AttributeDict(payment_data)

    def get_channels(self):
        return _make_request(f"{self.raiden.api_root_url}/channels")

    def get_new_payments(self, channel: models.Channel) -> List[AttributeDict]:

        offset = channel.payments.count()
        events = _make_request(channel.payments_url, offset=offset)
        assert type(events) is list

        payments = [self._parse_payment(ev, channel) for ev in events]
        return [payment for payment in payments if payment is not None]

    def get_token_addresses(self):
        return _make_request(f"{self.raiden.api_root_url}/tokens")



from django.urls import path

from . import views

app_name = "ethereum_money"


urlpatterns = [
    path("", views.TokenListView.as_view(), name="token-list"),
    path("token/<str:address>", views.TokenView.as_view(), name="token-detail"),
]



from django.urls import path
from rest_framework.routers import SimpleRouter


from . import views
from .consumers import CheckoutConsumer, NotificationConsumer

app_name = "hub20"


router = SimpleRouter(trailing_slash=False)
router.register("stores", views.StoreViewSet, basename="store")
router.register("checkout", views.CheckoutViewSet, basename="checkout")


urlpatterns = [
    path("balances", views.TokenBalanceListView.as_view(), name="balance-list"),
    path("balance/<str:code>", views.TokenBalanceView.as_view(), name="balance-detail"),
    path("payment/orders", views.PaymentOrderListView.as_view(), name="payment-order-list"),
    path("payment/order/<int:pk>", views.PaymentOrderView.as_view(), name="payment-order-detail"),
    path("transfers", views.TransferListView.as_view(), name="transfer-list"),
    path("transfers/transfer/<int:pk>", views.TransferView.as_view(), name="transfer-detail"),
] + router.urls


consumer_patterns = [
    path("checkout/<uuid:pk>", CheckoutConsumer),
    path("events", NotificationConsumer),
]



import factory
from factory import fuzzy

from hub20.apps.blockchain.factories import (
    TEST_CHAIN_ID,
    EthereumProvider,
    TransactionFactory,
    TransactionLogFactory,
)
from hub20.apps.blockchain.models import Transaction
from hub20.apps.ethereum_money import get_ethereum_account_model
from hub20.apps.ethereum_money.models import EthereumToken, EthereumTokenAmount

factory.Faker.add_provider(EthereumProvider)


EthereumAccount = get_ethereum_account_model()


class EthereumAccountFactory(factory.django.DjangoModelFactory):
    address = factory.Faker("ethereum_address")

    class Meta:
        model = EthereumAccount


class EthereumCurrencyFactory(factory.django.DjangoModelFactory):
    chain = TEST_CHAIN_ID


class ETHFactory(EthereumCurrencyFactory):
    name = fuzzy.FuzzyChoice(choices=["Ethereum"])
    code = fuzzy.FuzzyChoice(choices=["ETH"])

    class Meta:
        model = EthereumToken
        django_get_or_create = ("chain", "name")


class Erc20TokenFactory(EthereumCurrencyFactory):
    name = factory.Sequence(lambda n: f"ERC20 Token #{n:03}")
    code = factory.Sequence(lambda n: f"TOK#{n:03}")
    address = factory.Faker("ethereum_address")

    class Meta:
        model = EthereumToken


class EthereumTokenValueModelFactory(factory.django.DjangoModelFactory):
    amount = fuzzy.FuzzyDecimal(0, 10, precision=6)
    currency = factory.SubFactory(ETHFactory)


class Erc20TokenValueModelFactory(factory.django.DjangoModelFactory):
    amount = fuzzy.FuzzyDecimal(0, 1000, precision=8)
    currency = factory.SubFactory(Erc20TokenFactory)


class Erc20TokenAmountFactory(factory.Factory):
    amount = fuzzy.FuzzyDecimal(0, 10, precision=6)
    currency = factory.SubFactory(Erc20TokenFactory)

    class Meta:
        model = EthereumTokenAmount


class ETHAmountFactory(factory.Factory):
    amount = fuzzy.FuzzyDecimal(0, 10, precision=6)
    currency = factory.SubFactory(ETHFactory)

    class Meta:
        model = EthereumTokenAmount


class Erc20TransferFactory(TransactionFactory):
    log = factory.RelatedFactory(TransactionLogFactory, "transaction")

    class Meta:
        model = Transaction


__all__ = [
    "EthereumAccountFactory",
    "ETHFactory",
    "Erc20TokenFactory",
    "EthereumTokenValueModelFactory",
    "Erc20TokenValueModelFactory",
    "Erc20TokenAmountFactory",
    "ETHAmountFactory",
    "Erc20TransferFactory",
]



import factory

from hub20.apps.core import models

from .base import UserFactory
from .payments import Erc20TokenPaymentOrderFactory


class StoreFactory(factory.django.DjangoModelFactory):
    owner = factory.SubFactory(UserFactory)
    name = factory.Sequence(lambda n: f"Store #{n:02}")
    url = factory.Sequence(lambda n: f"https://store#{n:02}.example.com")

    @factory.post_generation
    def accepted_currencies(self, create, currencies, **kw):
        if not create:
            return

        if currencies:
            for currency in currencies:
                self.accepted_currencies.add(currency)

    class Meta:
        model = models.Store


class CheckoutFactory(factory.django.DjangoModelFactory):
    store = factory.SubFactory(StoreFactory, accepted_currencies=[])
    external_identifier = factory.fuzzy.FuzzyText(length=30, prefix="checkout-")
    payment_order = factory.SubFactory(
        Erc20TokenPaymentOrderFactory, user=factory.SelfAttribute("..store.owner"),
    )

    class Meta:
        model = models.Checkout


__all__ = ["StoreFactory", "CheckoutFactory"]



import datetime
import logging
from urllib.parse import urlparse

from django.conf import settings
from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.db.models import Avg, Max
from django.db.transaction import atomic
from django.utils import timezone
from web3 import Web3
from web3.middleware import geth_poa_middleware
from web3.providers import HTTPProvider, IPCProvider, WebsocketProvider

from .app_settings import CHAIN_ID, START_BLOCK_NUMBER
from .choices import ETHEREUM_CHAINS
from .fields import EthereumAddressField, HexField, Uint256Field

logger = logging.getLogger(__name__)


def database_history_gas_price_strategy(w3: Web3, transaction_params=None) -> int:
    BLOCK_HISTORY_SIZE = 100
    chain_id = int(w3.net.version)
    current_block_number = w3.eth.blockNumber
    default_price = Web3.toWei(1.2, "gwei")

    txs = Transaction.objects.filter(
        block__chain=chain_id, block__number__gte=current_block_number - BLOCK_HISTORY_SIZE
    )
    return int(txs.aggregate(avg_price=Avg("gas_price")).get("avg_price")) or default_price


def make_web3():
    web3_endpoint = settings.WEB3_PROVIDER_URI
    provider_class = {
        "http": HTTPProvider,
        "https": HTTPProvider,
        "ws": WebsocketProvider,
        "wss": WebsocketProvider,
    }.get(urlparse(web3_endpoint).scheme, IPCProvider)

    w3 = Web3(provider_class(web3_endpoint))
    w3.middleware_onion.inject(geth_poa_middleware, layer=0)
    w3.eth.setGasPriceStrategy(database_history_gas_price_strategy)

    chain_id = int(w3.net.version)
    assert chain_id == CHAIN_ID, f"web3 connected to network {chain_id}, expected {CHAIN_ID}"
    return w3


class Block(models.Model):
    hash = HexField(max_length=64, primary_key=True)
    chain = models.PositiveIntegerField(choices=ETHEREUM_CHAINS, default=ETHEREUM_CHAINS.mainnet)
    number = models.PositiveIntegerField(db_index=True)
    timestamp = models.DateTimeField()
    parent_hash = HexField(max_length=64)
    uncle_hashes = ArrayField(HexField(max_length=64))

    def __str__(self) -> str:
        hash_hex = self.hash if type(self.hash) is str else self.hash.hex()
        return f"{hash_hex} #{self.number} {self.get_chain_display()}"

    @property
    def parent(self):
        return self.__class__.objects.filter(hash=self.parent_hash).first()

    @property
    def uncles(self):
        return self.__class__.objects.filter(hash__in=self.uncle_hashes)

    @property
    def confirmations(self) -> int:
        chain_blocks = self.__class__.objects.filter(chain=self.chain)
        height = chain_blocks.aggregate(height=Max("number")).get("height")
        return height - self.number

    @classmethod
    def make(cls, block_data, chain_id):
        block_time = datetime.datetime.fromtimestamp(block_data.timestamp)
        block, _ = cls.objects.update_or_create(
            chain=chain_id,
            hash=block_data.hash,
            defaults={
                "number": block_data.number,
                "timestamp": timezone.make_aware(block_time),
                "parent_hash": block_data.parentHash,
                "uncle_hashes": block_data.uncles,
            },
        )
        return block

    @classmethod
    @atomic()
    def make_all(cls, block_number, w3):
        logger.info(f"Saving block {block_number}")
        chain_id = int(w3.net.version)
        block_data = w3.eth.getBlock(block_number, full_transactions=True)

        block = cls.make(block_data, chain_id)

        for tx_data in block_data.transactions:
            transaction = Transaction.make(tx_data, block)
            tx_receipt = w3.eth.waitForTransactionReceipt(transaction.hash)

            for log_data in tx_receipt.logs:
                TransactionLog.make(log_data, transaction)

        return block

    @classmethod
    def fetch_by_hash(cls, block_hash, w3: Web3 = None):
        try:
            return cls.objects.get(hash=block_hash)
        except cls.DoesNotExist:
            if w3 is None:
                w3 = make_web3()

            chain_id = int(w3.net.version)

            block_data = w3.eth.getBlock(block_hash, full_transactions=True)
            return cls.make(block_data, chain_id)

    @classmethod
    def get_latest_block_number(cls, qs):
        return qs.aggregate(latest=Max("number")).get("latest") or START_BLOCK_NUMBER

    class Meta:
        unique_together = ("chain", "hash", "number")


class Transaction(models.Model):
    block = models.ForeignKey(Block, on_delete=models.CASCADE)
    hash = HexField(max_length=64, db_index=True)
    from_address = EthereumAddressField(db_index=True)
    to_address = EthereumAddressField(db_index=True, null=True)
    gas = Uint256Field()
    gas_price = Uint256Field()
    nonce = Uint256Field()
    index = Uint256Field()
    value = Uint256Field()
    data = models.TextField()

    @classmethod
    def make(cls, transaction_data, block: Block):
        tx, _ = cls.objects.get_or_create(
            hash=transaction_data.hash,
            block=block,
            defaults={
                "from_address": transaction_data["from"],
                "to_address": transaction_data.to,
                "gas": transaction_data.gas,
                "gas_price": transaction_data.gasPrice,
                "nonce": transaction_data.nonce,
                "index": transaction_data.transactionIndex,
                "value": transaction_data.value,
                "data": transaction_data.input,
            },
        )
        return tx

    @classmethod
    def fetch_by_hash(cls, transaction_hash: str, w3: Web3 = None):
        try:
            return cls.objects.get(hash=transaction_hash)
        except cls.DoesNotExist:
            pass

        if w3 is None:
            w3 = make_web3()

        tx_data = w3.eth.getTransaction(transaction_hash)
        block = Block.fetch_by_hash(tx_data.blockHash, w3=w3)

        return cls.objects.filter(block=block, hash=transaction_hash).first()

    def __str__(self) -> str:
        hash_hex = self.hash if type(self.hash) is str else self.hash.hex()
        return f"Tx {hash_hex}"


class TransactionLog(models.Model):
    transaction = models.ForeignKey(Transaction, on_delete=models.CASCADE, related_name="logs")
    index = models.SmallIntegerField()
    data = models.TextField()
    topics = ArrayField(models.TextField())

    @classmethod
    def make(cls, log_data, transaction: Transaction):
        tx_log, _ = cls.objects.get_or_create(
            index=log_data.logIndex,
            transaction=transaction,
            defaults={"data": log_data.data, "topics": [l.hex() for l in log_data.topics]},
        )
        return tx_log

    class Meta:
        unique_together = ("transaction", "index")


__all__ = ["Block", "Transaction", "TransactionLog"]



import pytest
from asgiref.sync import sync_to_async
from channels.routing import URLRouter
from channels.testing import WebsocketCommunicator
from eth_utils import is_0x_prefixed

from hub20.apps.blockchain.factories import TransactionFactory
from hub20.apps.core.api import consumer_patterns
from hub20.apps.core.choices import PAYMENT_METHODS
from hub20.apps.core.factories import CheckoutFactory
from hub20.apps.core.signals import blockchain_payment_sent
from hub20.apps.ethereum_money.models import EthereumToken

application = URLRouter(consumer_patterns)


@pytest.mark.asyncio
@pytest.mark.django_db(transaction=True)
async def test_checkout_consumer():
    checkout = CheckoutFactory()
    checkout.store.accepted_currencies.add(checkout.payment_order.currency)

    communicator = WebsocketCommunicator(application, f"checkout/{checkout.id}")
    ok, protocol_or_error = await communicator.connect()

    assert ok, "Failed to connect"

    tx = TransactionFactory()
    await sync_to_async(blockchain_payment_sent.send)(
        sender=EthereumToken,
        amount=checkout.payment_order.as_token_amount,
        recipient=checkout.payment_order.payment_method.wallet.address,
        transaction_hash=tx.hash,
    )

    response = await communicator.receive_json_from()
    assert "event" in response
    assert "voucher" in response
    assert "identifier" in response
    assert "payment_method" in response

    assert is_0x_prefixed(response["identifier"])

    assert response["payment_method"] == PAYMENT_METHODS.blockchain
    assert response["event"] == "payment.sent"


__all__ = ["test_checkout_consumer"]



import factory
from factory import fuzzy

from hub20.apps.blockchain.factories.providers import EthereumProvider
from hub20.apps.ethereum_money.factories import (
    ETHFactory,
    Erc20TokenFactory,
    EthereumAccountFactory,
)
from hub20.apps.core import models
from .base import UserFactory


factory.Faker.add_provider(EthereumProvider)


class PaymentOrderFactory(factory.django.DjangoModelFactory):
    user = factory.SubFactory(UserFactory)
    amount = fuzzy.FuzzyDecimal(0, 10, precision=6)

    class Meta:
        abstract = False
        model = models.PaymentOrder


class ETHPaymentOrderFactory(PaymentOrderFactory):
    currency = factory.SubFactory(ETHFactory)

    class Meta:
        model = models.PaymentOrder


class Erc20TokenPaymentOrderFactory(PaymentOrderFactory):
    currency = factory.SubFactory(Erc20TokenFactory)

    class Meta:
        model = models.PaymentOrder


class BlockchainPaymentFactory(factory.django.DjangoModelFactory):
    order = factory.SubFactory(ETHPaymentOrderFactory)
    account = factory.SubFactory(EthereumAccountFactory)
    amount = factory.LazyAttribute(lambda obj: obj.order.amount)
    currency = factory.LazyAttribute(lambda obj: obj.order.currency)

    class Meta:
        model = models.BlockchainPayment


class PendingBlockchainPaymentFactory(BlockchainPaymentFactory):
    transaction_hash = factory.Faker("hex64")


__all__ = [
    "ETHPaymentOrderFactory",
    "Erc20TokenPaymentOrderFactory",
    "BlockchainPaymentFactory",
    "PendingBlockchainPaymentFactory",
]



from typing import Optional
import logging

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.db import transaction
from model_utils.managers import InheritanceManager
from model_utils.models import TimeStampedModel, StatusModel

from hub20.apps.blockchain.fields import EthereumAddressField, HexField
from hub20.apps.blockchain.models import Transaction
from hub20.apps.ethereum_money.models import EthereumTokenValueModel
from hub20.apps.raiden.models import Channel, Payment
from hub20.apps.core.choices import TRANSFER_EVENT_TYPES
from hub20.apps.core.signals import transfer_executed, transfer_failed, transfer_confirmed
from .accounting import UserAccount, Wallet, UserReserve

logger = logging.getLogger(__name__)


class TransferError(Exception):
    pass


class TransferOperationError(Exception):
    pass


class Transfer(TimeStampedModel, EthereumTokenValueModel):
    sender = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.PROTECT, related_name="transfers_sent"
    )
    memo = models.TextField(null=True, blank=True)
    identifier = models.CharField(max_length=300, null=True, blank=True)
    objects = InheritanceManager()

    @property
    def status(self) -> Optional[str]:
        try:
            return self.events.latest().status
        except TransferEvent.DoesNotExist:
            return None

    @property
    def is_finalized(self) -> bool:
        return self.status in [
            TRANSFER_EVENT_TYPES.confirmed,
            TRANSFER_EVENT_TYPES.failed,
            TRANSFER_EVENT_TYPES.canceled,
        ]

    def _execute(self):
        pass

    def _make_reserve(self):
        pass

    def verify_conditions(self):

        transfer_amount = self.as_token_amount
        sender_account = UserAccount(self.sender)
        sender_balance = sender_account.get_balance(self.currency)

        if sender_balance < transfer_amount:
            raise TransferError("Insufficient balance")

    def execute(self):
        if self.is_finalized:
            logger.warning(f"{self} is already finalized as {self.status}")
            return

        try:
            self.verify_conditions()
            self._make_reserve()
            self._execute()
        except TransferError as exc:
            logger.info(f"{self} failed: {exc}")
            transfer_failed.send_robust(sender=Transfer, transfer=self, reason=str(exc))
        except Exception as exc:
            logger.exception(exc)


class InternalTransfer(Transfer):
    receiver = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.PROTECT, related_name="transfers_received"
    )

    @property
    def target(self):
        return self.receiver.username

    def clean(self):
        if self.sender == self.receiver:
            raise ValidationError("Sender and Receiver are the same")

    @transaction.atomic()
    def _execute(self):
        transfer_confirmed.send_robust(sender=InternalTransfer, transfer=self)


class ExternalTransfer(Transfer):
    recipient_address = EthereumAddressField(db_index=True)

    @property
    def target(self):
        return self.recipient_address

    @transaction.atomic()
    def _execute(self):
        transfer_amount = self.as_token_amount
        channel = Channel.select_for_transfer(self.recipient_address, transfer_amount)
        if channel:
            payment_id = channel.send(
                self.recipient_address, transfer_amount, payment_identifier=self.identifier
            )
            RaidenTransaction.objects.create(
                transfer=self, channel=channel, payment_identifier=payment_id
            )
        else:
            wallet = Wallet.select_for_transfer(transfer_amount)
            if wallet is None:
                raise TransferError("No channel nor wallet with funds to make transfer found")
            tx_hash = wallet.account.send(self.recipient_address, transfer_amount)
            BlockchainTransaction.objects.create(transfer=self, transaction_hash=tx_hash)
        transfer_executed.send_robust(sender=ExternalTransfer, transfer=self)

    @transaction.atomic()
    def _make_reserve(self):
        self.sender.balance_entries.create(amount=-self.amount, currency=self.currency)
        self.sender.reserves.update_or_create(
            usertransferreserve__transfer=self,
            defaults={"amount": self.amount, "currency": self.currency},
        )


class UserTransferReserve(UserReserve):
    transfer = models.OneToOneField(Transfer, on_delete=models.CASCADE, related_name="reserve")


class BlockchainTransaction(TimeStampedModel):
    transfer = models.OneToOneField(
        Transfer, on_delete=models.CASCADE, related_name="chain_transaction"
    )
    transaction_hash = HexField(max_length=64, unique=True, db_index=True)

    @property
    def transaction(self) -> Optional[Transaction]:
        return Transaction.objects.filter(hash=self.transaction_hash).first()


class RaidenTransaction(TimeStampedModel):
    transfer = models.OneToOneField(
        Transfer, on_delete=models.CASCADE, related_name="raiden_transaction"
    )
    channel = models.ForeignKey(Channel, on_delete=models.CASCADE)
    identifier = models.PositiveIntegerField()

    @property
    def payment(self) -> Optional[Payment]:
        return self.channel.payments.filter(identifier=self.identifier).first()

    class Meta:
        unique_together = ("channel", "identifier")


class TransferEvent(TimeStampedModel, StatusModel):
    STATUS = TRANSFER_EVENT_TYPES

    transfer = models.ForeignKey(Transfer, on_delete=models.CASCADE, related_name="events")

    class Meta:
        get_latest_by = "created"
        unique_together = ("transfer", "status")


__all__ = [
    "Transfer",
    "InternalTransfer",
    "ExternalTransfer",
    "TransferError",
    "UserTransferReserve",
    "BlockchainTransaction",
    "RaidenTransaction",
]



import logging
from concurrent.futures import TimeoutError
from typing import Optional

from celery import shared_task
from hub20.apps.blockchain.models import make_web3, Block, Transaction

from .models import TokenNetwork, TokenNetworkChannel

logger = logging.getLogger(__name__)


def _get_channel_from_event(token_network, event) -> Optional[TokenNetworkChannel]:
    event_name = event.event
    if event_name == "ChannelOpened":
        participants = (event.args.participant1, event.args.participant2)
        channel_identifier = event.args.channel_identifier
        channel, _ = token_network.channels.get_or_create(
            identifier=channel_identifier, participant_addresses=participants
        )
        return channel

    if event_name == "ChannelClosed":
        channel_identifier = event.args.channel_identifier
        return token_network.channels.filter(identifier=channel_identifier).first()


def process_events(token_network, event_filter, w3):
    try:
        for event in event_filter.get_all_entries():
            logger.info(f"Processing event {event.event} at {event.transactionHash.hex()}")
            tx = Transaction.fetch_by_hash(event.transactionHash, w3=w3)
            if not tx:
                logger.warning(f"Transaction {event.transactionHash} could not be synced")

            channel = _get_channel_from_event(token_network, event)
            if channel:
                token_network.events.get_or_create(
                    channel=channel, transaction=tx, name=event.event
                )
            else:
                logger.warning(f"Failed to find channel related to event {event}")
    except TimeoutError:
        logger.error(f"Timed-out while getting events for {token_network}")
    except Exception as exc:
        logger.error(f"Failed to get events for {token_network}: {exc}")


@shared_task
def sync_token_network_events():
    w3 = make_web3()
    for token_network in TokenNetwork.objects.all():
        token_network_contract = token_network.get_contract(w3=w3)
        event_blocks = Block.objects.filter(
            transaction__tokennetworkchannelevent__channel__token_network=token_network
        )
        from_block = Block.get_latest_block_number(event_blocks) + 1

        logger.info(f"Fetching {token_network.token} events since block {from_block}")

        channel_open_filter = token_network_contract.events.ChannelOpened.createFilter(
            fromBlock=from_block
        )
        channel_closed_filter = token_network_contract.events.ChannelClosed.createFilter(
            fromBlock=from_block
        )

        process_events(token_network, channel_open_filter, w3)
        process_events(token_network, channel_closed_filter, w3)



from django.contrib import admin

from . import models


@admin.register(models.EthereumToken)
class TokenAdmin(admin.ModelAdmin):
    search_fields = ["code", "name", "address"]
    list_display = ["code", "name", "address", "chain"]
    list_filter = ["chain"]



from typing import Optional

from django.contrib import admin
from django.http import HttpRequest

from . import models


@admin.register(models.Wallet)
class WalletAdmin(admin.ModelAdmin):
    exclude = ["private_key"]

    def has_add_permission(
        self, request: HttpRequest, obj: Optional[models.Wallet] = None
    ) -> bool:
        return False

    def has_change_permission(
        self, request: HttpRequest, obj: Optional[models.Wallet] = None
    ) -> bool:
        return False

    def has_delete_permission(
        self, request: HttpRequest, obj: Optional[models.Wallet] = None
    ) -> bool:
        return False



from django.db.models.query import QuerySet
from django.http import Http404
from django.shortcuts import get_object_or_404
from eth_utils import is_address
from rest_framework import generics

from hub20.apps.blockchain.app_settings import CHAIN_ID

from . import models, serializers
from .app_settings import TRACKED_TOKENS


class TokenListView(generics.ListAPIView):
    serializer_class = serializers.HyperlinkedEthereumTokenSerializer

    def get_queryset(self) -> QuerySet:
        return models.EthereumToken.objects.filter(chain=CHAIN_ID, address__in=TRACKED_TOKENS)


class TokenView(generics.RetrieveAPIView):
    serializer_class = serializers.HyperlinkedEthereumTokenSerializer

    def get_object(self) -> models.EthereumToken:
        address = self.kwargs.get("address")

        if not is_address(address):
            raise Http404(f"{address} is not a valid token address")

        return get_object_or_404(models.EthereumToken, chain=CHAIN_ID, address=address)



from typing import List, Optional

from django.db.models.query import QuerySet
from django.http import Http404
from django.shortcuts import get_object_or_404
from rest_framework import generics
from rest_framework.mixins import CreateModelMixin, DestroyModelMixin, RetrieveModelMixin
from rest_framework.permissions import AllowAny, IsAuthenticated
from rest_framework.serializers import Serializer
from rest_framework.viewsets import GenericViewSet, ModelViewSet

from hub20.apps.blockchain.app_settings import CHAIN_ID
from hub20.apps.ethereum_money.models import EthereumToken, EthereumTokenAmount

from . import models, serializers


class ReadWriteSerializerMixin(generics.GenericAPIView):
    """
    Overrides get_serializer_class to choose the read serializer
    for GET requests and the write serializer for POST requests.

    Set read_serializer_class and write_serializer_class attributes on a
    generic APIView
    """

    read_serializer_class: Optional[Serializer] = None
    write_serializer_class: Optional[Serializer] = None

    def get_serializer_class(self) -> Serializer:
        if self.request.method in ["POST", "PUT", "PATCH"]:
            return self.get_write_serializer_class()
        return self.get_read_serializer_class()

    def get_read_serializer_class(self) -> Serializer:
        assert self.read_serializer_class is not None, (
            "'%s' should either include a `read_serializer_class` attribute,"
            "or override the `get_read_serializer_class()` method." % self.__class__.__name__
        )
        return self.read_serializer_class

    def get_write_serializer_class(self) -> Serializer:
        assert self.write_serializer_class is not None, (
            "'%s' should either include a `write_serializer_class` attribute,"
            "or override the `get_write_serializer_class()` method." % self.__class__.__name__
        )
        return self.write_serializer_class


class BasePaymentOrderView(ReadWriteSerializerMixin):
    read_serializer_class = serializers.PaymentOrderReadSerializer
    write_serializer_class = serializers.PaymentOrderSerializer


class PaymentOrderListView(BasePaymentOrderView, generics.ListCreateAPIView):
    permission_classes = (IsAuthenticated,)

    def get_queryset(self) -> QuerySet:
        return self.request.user.paymentorder_set.all()


class PaymentOrderView(BasePaymentOrderView, generics.RetrieveDestroyAPIView):
    permission_classes = (IsAuthenticated,)

    def get_object(self) -> models.PaymentOrder:
        return get_object_or_404(
            models.PaymentOrder, pk=self.kwargs.get("pk"), user=self.request.user
        )

    def destroy(self, request, *args, **kw):
        obj = self.get_object()
        obj.cancel()
        return self.retrieve(request, *args, **kw)


class TransferListView(generics.ListCreateAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.TransferSerializer

    def get_queryset(self) -> QuerySet:
        return self.request.user.transfers_sent.all()


class TransferView(generics.RetrieveAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.TransferSerializer

    def get_object(self):
        try:
            return models.Transfer.objects.get_subclass(
                pk=self.kwargs.get("pk"), sender=self.request.user
            )
        except models.Transfer.DoesNotExist:
            raise Http404


class TokenBalanceListView(generics.ListAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.TokenBalanceSerializer

    def get_queryset(self) -> List[EthereumTokenAmount]:
        return models.UserAccount(self.request.user).get_balances(chain_id=CHAIN_ID)


class TokenBalanceView(generics.RetrieveAPIView):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.TokenBalanceSerializer

    def get_object(self) -> EthereumTokenAmount:
        user_account = models.UserAccount(self.request.user)
        token = get_object_or_404(EthereumToken, code=self.kwargs["code"], chain=CHAIN_ID)
        return user_account.get_balance(token)


class CheckoutViewSet(GenericViewSet, CreateModelMixin, RetrieveModelMixin, DestroyModelMixin):
    permission_classes = (AllowAny,)
    serializer_class = serializers.HttpCheckoutSerializer
    lookup_value_regex = "[0-9a-f-]{36}"

    def get_queryset(self):
        return models.Checkout.objects.all()

    def get_object(self):
        return get_object_or_404(models.Checkout, id=self.kwargs["pk"])

    def destroy(self, request, pk=None):
        obj = self.get_object()
        obj.payment_order.cancel()
        return self.retrieve(request, pk=pk)


class StoreViewSet(ModelViewSet):
    permission_classes = (IsAuthenticated,)
    serializer_class = serializers.StoreSerializer

    def get_permissions(self):
        perms = (AllowAny,) if self.action == "retrieve" else (IsAuthenticated,)
        return (perm() for perm in perms)

    def get_queryset(self) -> QuerySet:
        return models.Store.objects.all()

    def get_object(self, *args, **kw):
        return get_object_or_404(models.Store, id=self.kwargs["pk"])



# Generated by Django 2.2.9 on 2020-01-30 13:14

import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import hub20.apps.blockchain.fields
import hub20.apps.ethereum_money.models
import model_utils.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('blockchain', '0001_initial'),
        ('ethereum_money', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='TokenNetwork',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('address', hub20.apps.blockchain.fields.EthereumAddressField()),
                ('token', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='ethereum_money.EthereumToken')),
            ],
        ),
        migrations.CreateModel(
            name='TokenNetworkChannel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identifier', models.PositiveIntegerField()),
                ('participant_addresses', django.contrib.postgres.fields.ArrayField(base_field=hub20.apps.blockchain.fields.EthereumAddressField(), size=2)),
                ('token_network', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='channels', to='raiden.TokenNetwork')),
            ],
        ),
        migrations.CreateModel(
            name='TokenNetworkChannelStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', model_utils.fields.StatusField(choices=[('open', 'open'), ('settling', 'settling'), ('settled', 'settled'), ('unusable', 'unusable'), ('closed', 'closed'), ('closing', 'closing')], default='open', max_length=100, no_check_for_status=True, verbose_name='status')),
                ('status_changed', model_utils.fields.MonitorField(default=django.utils.timezone.now, monitor='status', verbose_name='status changed')),
                ('channel', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='status', to='raiden.TokenNetworkChannel')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Raiden',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('address', hub20.apps.blockchain.fields.EthereumAddressField(db_index=True, unique=True)),
                ('url', models.URLField(help_text='Root URL of server (without api/v1)')),
                ('token_networks', models.ManyToManyField(blank=True, to='raiden.TokenNetwork')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Channel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', model_utils.fields.StatusField(choices=[('open', 'open'), ('settling', 'settling'), ('settled', 'settled'), ('unusable', 'unusable'), ('closed', 'closed'), ('closing', 'closing')], default='open', max_length=100, no_check_for_status=True, verbose_name='status')),
                ('status_changed', model_utils.fields.MonitorField(default=django.utils.timezone.now, monitor='status', verbose_name='status changed')),
                ('identifier', models.PositiveIntegerField()),
                ('partner_address', hub20.apps.blockchain.fields.EthereumAddressField(db_index=True)),
                ('balance', hub20.apps.ethereum_money.models.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('total_deposit', hub20.apps.ethereum_money.models.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('total_withdraw', hub20.apps.ethereum_money.models.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('raiden', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='channels', to='raiden.Raiden')),
                ('token_network', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='raiden.TokenNetwork')),
            ],
            options={
                'unique_together': {('raiden', 'token_network', 'identifier'), ('raiden', 'token_network', 'partner_address')},
            },
        ),
        migrations.CreateModel(
            name='TokenNetworkChannelEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, max_length=32)),
                ('channel', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='raiden.TokenNetworkChannel')),
                ('transaction', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='blockchain.Transaction')),
            ],
            options={
                'unique_together': {('channel', 'transaction')},
            },
        ),
        migrations.CreateModel(
            name='Payment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', hub20.apps.ethereum_money.models.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('timestamp', models.DateTimeField()),
                ('identifier', models.BigIntegerField()),
                ('sender_address', hub20.apps.blockchain.fields.EthereumAddressField()),
                ('receiver_address', hub20.apps.blockchain.fields.EthereumAddressField()),
                ('channel', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='payments', to='raiden.Channel')),
            ],
            options={
                'unique_together': {('channel', 'timestamp', 'sender_address', 'receiver_address')},
            },
        ),
    ]



import copy
from typing import Dict

from django.contrib.auth import get_user_model
from django.db import transaction
from ipware import get_client_ip
from rest_framework import serializers
from rest_framework.reverse import reverse

from hub20.apps.blockchain.serializers import EthereumAddressField, HexadecimalField
from hub20.apps.ethereum_money.serializers import (
    CurrencyRelatedField,
    EthereumTokenSerializer,
    TokenValueField,
)

from . import models

User = get_user_model()


class UserRelatedField(serializers.SlugRelatedField):
    queryset = User.objects.filter(is_active=True)

    def __init__(self, *args, **kw):
        kw.setdefault("slug_field", "username")
        super().__init__(*args, **kw)


class TokenBalanceSerializer(serializers.Serializer):
    url = serializers.SerializerMethodField()
    amount = TokenValueField(read_only=True)
    currency = EthereumTokenSerializer(read_only=True)
    balance = serializers.CharField(source="formatted", read_only=True)

    def get_url(self, obj):
        return reverse(
            "hub20:balance-detail",
            kwargs={"code": obj.currency.code},
            request=self.context.get("request"),
        )


class TransferSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="hub20:transfer-detail")
    address = EthereumAddressField(write_only=True, required=False)
    recipient = UserRelatedField(write_only=True, required=False, allow_null=True)
    token = CurrencyRelatedField()
    target = serializers.CharField(read_only=True)
    status = serializers.CharField(read_only=True)

    def validate_recipient(self, value):
        request = self.context["request"]
        if value == request.user:
            raise serializers.ValidationError("You can not make a transfer to yourself")
        return value

    def validate(self, data):
        address = data.get("address")
        recipient = data.get("recipient")
        request = self.context["request"]

        if not address and not recipient:
            raise serializers.ValidationError(
                "Either one of recipient or address must be provided"
            )
        if address and recipient:
            raise serializers.ValidationError(
                "Choose recipient by address or username, but not both at the same time"
            )

        transfer_data = copy.copy(data)
        del transfer_data["address"]
        del transfer_data["recipient"]

        if recipient is not None:
            transfer_class = models.InternalTransfer
            transfer_data["receiver"] = recipient
        else:
            transfer_class = models.ExternalTransfer
            transfer_data["recipient_address"] = address

        transfer = transfer_class(sender=request.user, **transfer_data)

        try:
            transfer.verify_conditions()
        except models.TransferError as exc:
            raise serializers.ValidationError(str(exc))

        return transfer_data

    def create(self, validated_data):
        transfer_class = (
            models.InternalTransfer if "recipient" in validated_data else models.ExternalTransfer
        )
        request = self.context["request"]

        return transfer_class.objects.create(sender=request.user, **validated_data)

    class Meta:
        model = models.Transfer
        fields = (
            "url",
            "address",
            "recipient",
            "amount",
            "token",
            "memo",
            "identifier",
            "status",
            "target",
        )
        read_only_fields = ("recipient", "status", "target")


class PaymentOrderMethodSerializer(serializers.ModelSerializer):
    blockchain = EthereumAddressField(source="wallet.address", read_only=True)
    raiden = EthereumAddressField(source="raiden.address", read_only=True)
    identifier = serializers.IntegerField(read_only=True)
    expiration_time = serializers.DateTimeField(read_only=True)

    class Meta:
        model = models.PaymentOrderMethod
        fields = ("blockchain", "raiden", "identifier", "expiration_time")
        read_only_fields = ("blockchain", "raiden", "identifier", "expiration_time")


class PaymentSerializer(serializers.ModelSerializer):
    currency = EthereumTokenSerializer()
    identifier = serializers.CharField()
    confirmed = serializers.BooleanField(source="is_confirmed")

    class Meta:
        model = models.Payment
        fields = ("created", "currency", "amount", "identifier", "route", "confirmed")
        read_only_fields = ("created", "currency", "amount", "identifier", "route", "confirmed")


class InternalPaymentSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.InternalPayment
        fields = PaymentSerializer.Meta.fields + ("user", "memo")
        read_only_fields = PaymentSerializer.Meta.read_only_fields + ("user", "memo")


class BlockchainPaymentSerializer(PaymentSerializer):
    transaction = HexadecimalField(read_only=True, source="transaction.hash")

    class Meta:
        model = models.BlockchainPayment
        fields = PaymentSerializer.Meta.fields + ("transaction",)
        read_only_fields = PaymentSerializer.Meta.read_only_fields + ("transaction",)


class RaidenPaymentSerializer(PaymentSerializer):
    raiden = serializers.CharField(source="payment.channel.raiden.address")

    class Meta:
        model = models.RaidenPayment
        fields = PaymentSerializer.Meta.fields + ("raiden",)
        read_only_fields = PaymentSerializer.Meta.read_only_fields + ("raiden",)


class PaymentOrderSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="hub20:payment-order-detail")
    currency = CurrencyRelatedField()
    payment_method = PaymentOrderMethodSerializer(read_only=True)
    payments = serializers.SerializerMethodField()

    def create(self, validated_data: Dict) -> models.PaymentOrder:
        request = self.context["request"]
        with transaction.atomic():
            return models.PaymentOrder.objects.create(user=request.user, **validated_data)

    def get_payments(self, obj):
        def get_payment_serializer(payment):
            return {
                models.InternalPayment: InternalPaymentSerializer,
                models.BlockchainPayment: BlockchainPaymentSerializer,
                models.RaidenPayment: RaidenPaymentSerializer,
            }.get(type(payment), PaymentSerializer)(payment, context=self.context)

        return [get_payment_serializer(payment).data for payment in obj.payments]

    class Meta:
        model = models.PaymentOrder
        fields = ["url", "amount", "currency", "created", "payment_method", "payments", "status"]
        read_only_fields = ["payment_method", "status", "created"]


class PaymentOrderReadSerializer(PaymentOrderSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="hub20:payment-order-detail")
    currency = EthereumTokenSerializer(read_only=True)


class CheckoutSerializer(serializers.ModelSerializer):
    store = serializers.PrimaryKeyRelatedField(queryset=models.Store.objects.all())
    token = CurrencyRelatedField(source="payment_order.currency")
    amount = TokenValueField(source="payment_order.amount")
    status = serializers.CharField(source="payment_order.status", read_only=True)
    payment_method = PaymentOrderMethodSerializer(
        source="payment_order.payment_method", read_only=True
    )
    payments = PaymentSerializer(many=True, source="payment_order.payments", read_only=True)
    voucher = serializers.SerializerMethodField()

    def validate(self, data):
        token = data["payment_order"]["currency"]
        store = data["store"]
        if token not in store.accepted_currencies.all():
            raise serializers.ValidationError(f"{token.code} is not accepted at {store.name}")

        return data

    def create(self, validated_data):
        request = self.context.get("request")
        client_ip, _ = get_client_ip(request)
        store = validated_data["store"]
        with transaction.atomic():
            payment_order = models.PaymentOrder.objects.create(
                user=store.owner, **validated_data["payment_order"]
            )

            return models.Checkout.objects.create(
                store=store,
                external_identifier=validated_data["external_identifier"],
                payment_order=payment_order,
                requester_ip=client_ip,
            )

    def get_voucher(self, obj):
        return obj.issue_voucher()

    class Meta:
        model = models.Checkout
        fields = (
            "id",
            "created",
            "store",
            "external_identifier",
            "token",
            "amount",
            "payment_method",
            "payments",
            "status",
            "voucher",
        )
        read_only_fields = ("id",)


class HttpCheckoutSerializer(CheckoutSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="hub20:checkout-detail")

    class Meta:
        model = models.Checkout
        fields = ("url",) + CheckoutSerializer.Meta.fields
        read_only_fields = CheckoutSerializer.Meta.read_only_fields


class StoreSerializer(serializers.ModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="hub20:store-detail")
    site_url = serializers.URLField(source="url")
    public_key = serializers.CharField(source="rsa.public_key_pem", read_only=True)
    accepted_currencies = CurrencyRelatedField(many=True)

    def create(self, validated_data):
        request = self.context.get("request")
        currencies = validated_data.pop("accepted_currencies", [])
        store = models.Store.objects.create(owner=request.user, **validated_data)
        store.accepted_currencies.set(currencies)
        return store

    class Meta:
        model = models.Store
        fields = ("id", "url", "name", "site_url", "public_key", "accepted_currencies")
        read_only_fields = ("id", "public_key")



import os

from celery import Celery
from celery.schedules import crontab


class Hub20CeleryConfig:
    broker_url = "memory" if "HUB20_TEST" in os.environ else os.getenv("HUB20_BROKER_URL")
    broker_use_ssl = "HUB20_BROKER_USE_SSL" in os.environ
    beat_schedule = {
        "sync-token-network-events": {
            "task": "hub20.apps.raiden.tasks.sync_token_network_events",
            "schedule": crontab(minute="*/10"),
        }
    }
    task_always_eager = "HUB20_TEST" in os.environ
    task_eager_propagates = "HUB20_TEST" in os.environ


app = Celery()
app.config_from_object(Hub20CeleryConfig)
app.autodiscover_tasks()



import logging
import random
from typing import List, Optional, TypeVar

from django.conf import settings
from django.contrib.sites.models import Site
from django.db import models
from model_utils.models import TimeStampedModel
from model_utils.managers import InheritanceManager

from hub20.apps.ethereum_money import get_ethereum_account_model
from hub20.apps.ethereum_money.models import (
    EthereumTokenValueModel,
    EthereumTokenAmount,
    EthereumToken,
    get_max_fee,
)
from hub20.apps.raiden.models import Channel


logger = logging.getLogger(__name__)
Wallet_T = TypeVar("Wallet_T", bound="Wallet")


EthereumAccount = get_ethereum_account_model()


class Wallet(models.Model):
    account = models.OneToOneField(EthereumAccount, on_delete=models.CASCADE)
    objects = InheritanceManager()

    @property
    def address(self):
        return self.account.address

    def __str__(self) -> str:
        return self.address

    def get_balance(self, currency: EthereumToken) -> EthereumTokenAmount:
        return EthereumTokenAmount.aggregated(
            self.account.balance_entries.all(), currency=currency
        )

    def get_balances(self, chain_id: int) -> List[EthereumTokenAmount]:
        return [self.get_balance(token) for token in EthereumToken.objects.filter(chain=chain_id)]

    @classmethod
    def select_for_transfer(cls, transfer_amount: EthereumTokenAmount) -> Optional[Wallet_T]:
        max_fee_amount: EthereumTokenAmount = get_max_fee()
        assert max_fee_amount.is_ETH

        ETH = max_fee_amount.currency

        eth_required = max_fee_amount
        token_required = EthereumTokenAmount(
            amount=transfer_amount.amount, currency=transfer_amount.currency
        )
        wallets = cls.objects.all()

        if transfer_amount.is_ETH:
            token_required += eth_required
            funded_wallets = [w for w in wallets if w.get_balance(ETH) >= token_required]
        else:
            funded_wallets = [
                wallet
                for wallet in wallets
                if wallet.get_balance(token_required.currency) >= token_required
                and wallet.get_balance(ETH) >= eth_required
            ]

        try:
            return random.choice(funded_wallets)
        except IndexError:
            return None

    @classmethod
    def generate(cls):
        return cls.objects.create(account=EthereumAccount.generate())


class UserBalanceEntry(TimeStampedModel, EthereumTokenValueModel):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name="balance_entries"
    )


class UserReserve(TimeStampedModel, EthereumTokenValueModel):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name="reserves"
    )


class UserAccount:
    def __init__(self, user):
        self.user = user

    def get_balance(self, currency: EthereumToken) -> EthereumTokenAmount:
        return EthereumTokenAmount.aggregated(self.user.balance_entries.all(), currency=currency)

    def get_balances(self, chain_id: int) -> List[EthereumTokenAmount]:
        return [self.get_balance(token) for token in EthereumToken.objects.filter(chain=chain_id)]


class HubSite(Site):
    def get_funds(self, currency: EthereumToken) -> EthereumTokenAmount:
        wallet_funds = sum([wallet.get_balance(currency) for wallet in Wallet.objects.all()])
        channel_funds = sum([c.balance_amount for c in Channel.objects.filter(currency=currency)])
        return wallet_funds + channel_funds

    class Meta:
        proxy = True


__all__ = [
    "HubSite",
    "UserBalanceEntry",
    "UserAccount",
    "UserReserve",
    "Wallet",
]



import datetime
import logging

from django.db.models.signals import post_save
from django.dispatch import receiver

from hub20.apps.blockchain.models import Block, Transaction
from hub20.apps.core import tasks
from hub20.apps.core.app_settings import PAYMENT_SETTINGS, TRANSFER_SETTINGS
from hub20.apps.core.choices import PAYMENT_EVENT_TYPES, PAYMENT_METHODS, TRANSFER_EVENT_TYPES
from hub20.apps.core.models import (
    BlockchainPayment,
    Checkout,
    ExternalTransfer,
    InternalPayment,
    InternalTransfer,
    PaymentOrder,
    PaymentOrderEvent,
    PaymentOrderMethod,
    RaidenPayment,
    Store,
    StoreRSAKeyPair,
    Transfer,
    Wallet,
)
from hub20.apps.core.signals import (
    blockchain_payment_sent,
    order_canceled,
    order_paid,
    payment_confirmed,
    payment_received,
    transfer_confirmed,
    transfer_executed,
    transfer_failed,
    transfer_scheduled,
)
from hub20.apps.ethereum_money.models import EthereumToken
from hub20.apps.ethereum_money.signals import account_deposit_received
from hub20.apps.raiden.models import Payment as RaidenPaymentEvent, Raiden
from hub20.apps.raiden.signals import raiden_payment_received

logger = logging.getLogger(__name__)


@receiver(account_deposit_received, sender=Transaction)
def on_account_deposit_check_blockchain_payments(sender, **kw):
    account = kw["account"]
    transaction = kw["transaction"]
    amount = kw["amount"]
    order = PaymentOrder.objects.filter(payment_method__wallet__account=account).first()

    if not order:
        return

    payment = BlockchainPayment.objects.create(
        order=order, amount=amount.amount, currency=amount.currency, transaction=transaction
    )
    payment_received.send(sender=BlockchainPayment, payment=payment)


@receiver(blockchain_payment_sent, sender=EthereumToken)
def on_blockchain_payment_sent_maybe_publish_checkout(sender, **kw):
    recipient_address = kw["recipient"]
    payment_amount = kw["amount"]
    tx_hash = kw["transaction_hash"]
    checkout = Checkout.objects.filter(
        payment_order__payment_method__wallet__account__address=recipient_address
    ).first()

    if not checkout:
        return

    if not checkout.payment_order.is_finalized:
        tasks.publish_checkout_event.delay(
            checkout.pk,
            event="payment.sent",
            amount=payment_amount.amount,
            token=payment_amount.currency.address,
            identifier=tx_hash,
            payment_method=PAYMENT_METHODS.blockchain,
        )


@receiver(raiden_payment_received, sender=RaidenPaymentEvent)
def on_raiden_payment_received_check_raiden_payments(sender, **kw):
    raiden_payment = kw["payment"]

    order = PaymentOrder.objects.filter(
        payment_method__identifier=raiden_payment.identifier,
        payment_method__raiden=raiden_payment.channel.raiden,
    ).first()

    if order is not None:
        amount = raiden_payment.as_token_amount
        payment = RaidenPayment.objects.create(
            order=order,
            amount=amount.amount,
            currency=raiden_payment.token,
            payment=raiden_payment,
        )
        payment_confirmed.send(sender=RaidenPayment, payment=payment)


@receiver(post_save, sender=PaymentOrder)
def on_order_created_set_payment_methods(sender, **kw):
    order = kw["instance"]
    if kw["created"]:
        unlocked_wallets = Wallet.objects.filter(paymentordermethod__isnull=True)
        wallet = unlocked_wallets.order_by("?").first() or Wallet.generate()

        raiden = Raiden.objects.filter(token_networks__token=order.currency).first()
        expiration_time = order.created + datetime.timedelta(
            seconds=PaymentOrderMethod.EXPIRATION_TIME
        )

        PaymentOrderMethod.objects.create(
            order=order, wallet=wallet, raiden=raiden, expiration_time=expiration_time
        )


@receiver(post_save, sender=PaymentOrder)
def on_payment_created_set_created_status(sender, **kw):
    payment = kw["instance"]
    if payment.created:
        payment.events.create(status=PAYMENT_EVENT_TYPES.requested)


@receiver(post_save, sender=PaymentOrderMethod)
def on_payment_order_method_created_schedule_expiration_task(sender, **kw):
    payment_method = kw["instance"]
    if kw["created"]:

        tasks.expire_payment_method.apply_async(
            args=(payment_method.id,), eta=payment_method.expiration_time
        )


@receiver(post_save, sender=PaymentOrderEvent)
def on_payment_event_created_send_order_paid_signal(sender, **kw):
    payment_event = kw["instance"]
    if payment_event.created and payment_event.status == PAYMENT_EVENT_TYPES.confirmed:
        order_paid.send(sender=PaymentOrder, payment_order=payment_event.order)


@receiver(post_save, sender=PaymentOrderEvent)
def on_payment_order_canceled_event_send_order_canceled_signal(sender, **kw):
    payment_event = kw["instance"]
    if payment_event.created and payment_event.status == PAYMENT_EVENT_TYPES.canceled:
        order_canceled.send(sender=PaymentOrder, payment_order=payment_event.order)


@receiver(post_save, sender=Block)
def on_block_added_check_confirmed_payments(sender, **kw):
    block = kw["instance"]
    created = kw["created"]

    block_number_to_confirm = block.number - PAYMENT_SETTINGS.minimum_confirmations
    if created and block_number_to_confirm >= 0:
        payments = BlockchainPayment.objects.all()

        for payment in payments.filter(
            transaction__block__number=block_number_to_confirm,
            transaction__block__chain=block.chain,
        ):
            logger.info(f"Confirming {payment}")
            payment_confirmed.send(sender=BlockchainPayment, payment=payment)


@receiver(post_save, sender=Block)
def on_block_added_check_confirmed_transfers(sender, **kw):
    block = kw["instance"]
    created = kw["created"]

    block_number_to_confirm = block.number - TRANSFER_SETTINGS.minimum_confirmations
    if created and block_number_to_confirm >= 0:
        transactions = Transaction.objects.filter(
            block__number=block_number_to_confirm, block__chain=block.chain
        )

        tx_hashes = transactions.values_list("hash", flat=True)
        transfers = ExternalTransfer.objects.all()
        for transfer in transfers.filter(chain_transaction__transaction_hash__in=tx_hashes):
            logger.info(f"Confirming {transfer}")
            transfer_confirmed.send(sender=ExternalTransfer, transfer=transfer)


@receiver(payment_received, sender=InternalPayment)
@receiver(payment_received, sender=BlockchainPayment)
def on_payment_received_update_status(sender, **kw):
    payment = kw["payment"]
    logger.info(f"Processing payment {payment} received")
    payment.order.update_status()
    payment.order.maybe_finalize()


@receiver(payment_received, sender=BlockchainPayment)
def on_blockchain_payment_received_maybe_publish_checkout(sender, **kw):
    payment = kw["payment"]

    checkouts = Checkout.objects.filter(payment_order__payment=payment)
    checkout_id = checkouts.values_list("id", flat=True).first()

    if checkout_id is None:
        return

    tasks.publish_checkout_event.delay(
        checkout_id,
        amount=payment.amount,
        token=payment.currency.address,
        identifier=payment.transaction.hash.hex(),
        payment_method=PAYMENT_METHODS.blockchain,
        event="payment.received",
    )


@receiver(payment_confirmed, sender=InternalPayment)
@receiver(payment_confirmed, sender=BlockchainPayment)
@receiver(payment_confirmed, sender=RaidenPayment)
def on_payment_confirmed_publish_checkout(sender, **kw):
    payment = kw["payment"]

    checkouts = Checkout.objects.filter(payment_order__payment=payment)
    checkout_id = checkouts.values_list("id", flat=True).first()

    payment_method = {
        InternalPayment: PAYMENT_METHODS.internal,
        BlockchainPayment: PAYMENT_METHODS.blockchain,
        RaidenPayment: PAYMENT_METHODS.raiden,
    }.get(sender)

    if checkout_id is None:
        return

    tasks.publish_checkout_event.delay(
        checkout_id,
        amount=payment.amount,
        token=payment.currency.address,
        event="payment.confirmed",
        payment_method=payment_method,
    )


@receiver(payment_confirmed, sender=BlockchainPayment)
@receiver(payment_confirmed, sender=RaidenPayment)
def on_payment_confirmed_finalize(sender, **kw):
    payment = kw["payment"]
    logger.info(f"Processing payment {payment} confirmed")
    payment.order.maybe_finalize()


@receiver(order_paid, sender=PaymentOrder)
def on_order_paid_credit_user(sender, **kw):
    order = kw["payment_order"]

    order.user.balance_entries.create(amount=order.amount, currency=order.currency)


@receiver(order_paid, sender=PaymentOrder)
@receiver(order_canceled, sender=PaymentOrder)
def on_order_finalized_free_payment_channels(sender, **kw):
    order = kw["payment_order"]

    try:
        order.payment_method.delete()
    except PaymentOrder.payment_method.RelatedObjectDoesNotExist:
        pass
    except Exception as exc:
        logger.exception(exc)


@receiver(post_save, sender=InternalTransfer)
@receiver(post_save, sender=ExternalTransfer)
def on_transfer_created_mark_transfer_scheduled(sender, **kw):
    transfer = kw["instance"]
    if kw["created"]:
        transfer.events.create(status=TRANSFER_EVENT_TYPES.scheduled)
        tasks.execute_transfer.delay(transfer.id)
        transfer_scheduled.send(sender=sender, transfer=transfer)


@receiver(transfer_failed, sender=Transfer)
def on_transfer_failed_mark_as_failed(sender, **kw):
    transfer = kw["transfer"]
    transfer.events.create(status=TRANSFER_EVENT_TYPES.failed)


@receiver(transfer_confirmed, sender=ExternalTransfer)
@receiver(transfer_confirmed, sender=InternalTransfer)
def on_transfer_confirmed_mark_as_confirmed(sender, **kw):
    transfer = kw["transfer"]
    transfer.events.create(status=TRANSFER_EVENT_TYPES.confirmed)


@receiver(transfer_executed, sender=ExternalTransfer)
def on_external_transfer_executed_mark_as_executed(sender, **kw):
    transfer = kw["transfer"]
    transfer.events.create(status=TRANSFER_EVENT_TYPES.executed)


@receiver(transfer_confirmed, sender=ExternalTransfer)
def on_external_transfer_confirmed_destroy_reserve(sender, **kw):
    transfer = kw["transfer"]
    try:
        transfer.reserve.delete()
    except Transfer.reserve.RelatedObjectDoesNotExist:
        pass
    except Exception as exc:
        logger.exception(exc)


@receiver(transfer_confirmed, sender=InternalTransfer)
def on_internal_transfer_confirmed_move_balances(sender, **kw):
    transfer = kw["transfer"]
    transfer.receiver.balance_entries.create(amount=transfer.amount, currency=transfer.currency)
    transfer.sender.balance_entries.create(amount=-transfer.amount, currency=transfer.currency)


@receiver(post_save, sender=Store)
def on_store_created_generate_key_pair(sender, **kw):
    store = kw["instance"]
    if kw["created"]:
        StoreRSAKeyPair.generate(store)


__all__ = [
    "on_account_deposit_check_blockchain_payments",
    "on_blockchain_payment_sent_maybe_publish_checkout",
    "on_blockchain_payment_received_maybe_publish_checkout",
    "on_block_added_check_confirmed_payments",
    "on_block_added_check_confirmed_transfers",
    "on_payment_created_set_created_status",
    "on_payment_order_method_created_schedule_expiration_task",
    "on_payment_event_created_send_order_paid_signal",
    "on_payment_received_update_status",
    "on_payment_confirmed_finalize",
    "on_payment_confirmed_publish_checkout",
    "on_payment_order_canceled_event_send_order_canceled_signal",
    "on_order_paid_credit_user",
    "on_order_finalized_free_payment_channels",
    "on_order_created_set_payment_methods",
    "on_transfer_created_mark_transfer_scheduled",
    "on_transfer_failed_mark_as_failed",
    "on_transfer_confirmed_mark_as_confirmed",
    "on_internal_transfer_confirmed_move_balances",
    "on_external_transfer_executed_mark_as_executed",
    "on_external_transfer_confirmed_destroy_reserve",
    "on_store_created_generate_key_pair",
]



from raiden_contracts.contract_manager import (
    ContractManager,
    contracts_precompiled_path,
    get_contracts_deployment_info,
)
from raiden_contracts.constants import CONTRACT_TOKEN_NETWORK_REGISTRY, CONTRACT_TOKEN_NETWORK
from web3 import Web3


def _get_contract(w3: Web3, contract_name: str):
    chain_id = int(w3.net.version)
    manager = ContractManager(contracts_precompiled_path())

    contract_data = get_contracts_deployment_info(chain_id)
    assert contract_data
    address = contract_data["contracts"][contract_name]["address"]

    abi = manager.get_contract_abi(contract_name)
    return w3.eth.contract(abi=abi, address=address)


def get_token_network_registry_contract(w3: Web3):
    return _get_contract(w3, CONTRACT_TOKEN_NETWORK_REGISTRY)


def get_token_network_contract(w3):
    return _get_contract(w3, CONTRACT_TOKEN_NETWORK)



import random

from django.conf import settings
from django.db import models
from django.db.models import Sum
from model_utils.managers import InheritanceManager
from model_utils.models import StatusModel, TimeStampedModel

from hub20.apps.blockchain.models import Transaction
from hub20.apps.core.app_settings import PAYMENT_SETTINGS
from hub20.apps.core.choices import PAYMENT_EVENT_TYPES
from hub20.apps.ethereum_money import get_ethereum_account_model
from hub20.apps.ethereum_money.models import EthereumTokenValueModel
from hub20.apps.raiden.models import Payment as RaidenPaymentEvent, Raiden

from .accounting import Wallet

EthereumAccount = get_ethereum_account_model()


def generate_payment_order_id():
    # Javascript can not handle numbers bigger than 2^53 - 1, so let's make
    # that our upper bound
    return random.randint(1, 2 ** 53 - 1)


class PaymentOrder(TimeStampedModel, EthereumTokenValueModel):

    STATUS = PAYMENT_EVENT_TYPES
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    @property
    def payments(self):
        return self.payment_set.filter(currency=self.currency).select_subclasses()

    @property
    def total_transferred(self):
        return self.payments.aggregate(total=Sum("amount")).get("total") or 0

    @property
    def total_confirmed(self):
        return sum([payment.amount for payment in self.payments if payment.is_confirmed])

    @property
    def due_amount(self):
        return max(0, self.amount - self.total_transferred)

    @property
    def status(self):
        try:
            return self.events.latest().status
        except PaymentOrderEvent.DoesNotExist:
            return None

    @property
    def is_finalized(self):
        return self.status not in [
            self.STATUS.requested,
            self.STATUS.partial,
            self.STATUS.expired,
            self.STATUS.received,
        ]

    def update_status(self):
        if self.is_finalized:
            return

        due_amount = self.due_amount
        total_transferred = self.total_transferred

        if total_transferred > 0 and due_amount > 0 and self.status != self.STATUS.partial:
            self.events.create(status=self.STATUS.partial)
        elif total_transferred > 0 and due_amount == 0 and self.status != self.STATUS.received:
            self.events.create(status=self.STATUS.received)
        elif self.total_transferred == 0 and self.status != self.STATUS.requested:
            self.events.create(status=self.STATUS.requested)

    def maybe_finalize(self):
        if self.is_finalized:
            return

        if self.total_confirmed >= self.amount:
            self.events.create(status=self.STATUS.confirmed)

    def cancel(self):
        if self.is_finalized:
            return

        self.events.create(status=self.STATUS.canceled)


class PaymentOrderMethod(TimeStampedModel):
    EXPIRATION_TIME = PAYMENT_SETTINGS.payment_lifetime

    order = models.OneToOneField(
        PaymentOrder, on_delete=models.CASCADE, related_name="payment_method"
    )
    expiration_time = models.DateTimeField()
    wallet = models.OneToOneField(Wallet, on_delete=models.CASCADE)
    raiden = models.ForeignKey(Raiden, null=True, on_delete=models.SET_NULL)
    identifier = models.BigIntegerField(default=generate_payment_order_id, unique=True)


class PaymentOrderEvent(TimeStampedModel, StatusModel):
    STATUS = PAYMENT_EVENT_TYPES
    order = models.ForeignKey(PaymentOrder, on_delete=models.PROTECT, related_name="events")

    class Meta:
        get_latest_by = "created"
        unique_together = ("order", "status")


class Payment(TimeStampedModel, EthereumTokenValueModel):
    order = models.ForeignKey(PaymentOrder, on_delete=models.PROTECT)
    objects = InheritanceManager()

    @property
    def is_confirmed(self):
        return True

    @property
    def route(self):
        return self.__class__.ROUTE


class InternalPayment(Payment):
    ROUTE = "internal"

    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)
    memo = models.TextField(null=True, blank=True)

    @property
    def identifier(self):
        return str(self.id)


class BlockchainPayment(Payment):
    ROUTE = "blockchain"

    transaction = models.OneToOneField(Transaction, on_delete=models.CASCADE)

    @property
    def is_confirmed(self):
        return self.transaction.block.confirmations >= PAYMENT_SETTINGS.minimum_confirmations

    @property
    def identifier(self):
        return str(self.transaction.hash)


class RaidenPayment(Payment):
    ROUTE = "raiden"

    payment = models.OneToOneField(RaidenPaymentEvent, on_delete=models.PROTECT)

    @property
    def identifier(self):
        return f"{self.payment.identifier}-{self.id}"


__all__ = [
    "PaymentOrder",
    "PaymentOrderMethod",
    "PaymentOrderEvent",
    "Payment",
    "InternalPayment",
    "BlockchainPayment",
    "RaidenPayment",
]



import logging
import uuid
from typing import Union

from asgiref.sync import async_to_sync
from channels.auth import get_user
from channels.generic.websocket import JsonWebsocketConsumer

from . import models

logger = logging.getLogger(__name__)


def accept_subprotocol(consumer):
    try:
        subprotocol = consumer.scope["subprotocols"][0]
        consumer.accept(subprotocol)
    except IndexError:
        consumer.accept()


class NotificationConsumer(JsonWebsocketConsumer):
    @classmethod
    def get_group_name(cls, user):
        return f"notifications.{user.username}"

    def connect(self):
        accept_subprotocol(self)
        user = async_to_sync(get_user)(self.scope)
        if not user.is_authenticated:
            self.close()

        group_name = self.__class__.get_group_name(user)
        async_to_sync(self.channel_layer.group_add)(group_name, self.channel_name)

    def notify_payment_status_change(self, payment_data):
        self.send_json(payment_data)


class CheckoutConsumer(JsonWebsocketConsumer):
    @classmethod
    def get_group_name(cls, checkout_id: Union[uuid.UUID, str]) -> str:
        uid = uuid.UUID(str(checkout_id))
        return f"checkout.{uid.hex}"

    def connect(self):
        checkout_id = self.scope["url_route"]["kwargs"].get("pk")
        checkout = models.Checkout.objects.filter(id=checkout_id).first()

        if not checkout:
            self.close()

        self.checkout = checkout
        group_name = self.__class__.get_group_name(checkout_id)
        async_to_sync(self.channel_layer.group_add)(group_name, self.channel_name)

        accept_subprotocol(self)

    def _refresh(self):
        self.checkout.refresh_from_db()

    def publish_payment_event(self, message):
        logger.info(f"Message received: {message}")
        self._refresh()
        message.pop("type", None)
        message["event"] = message.pop("event_name", None)
        message["voucher"] = self.checkout.issue_voucher()
        self.send_json(message)



from django.apps import AppConfig


class EthereumMoneyConfig(AppConfig):
    name = "hub20.apps.ethereum_money"

    def ready(self):
        from . import signals  # noqa
        from . import handlers  # noqa



import factory

from hub20.apps.blockchain.factories import EthereumProvider
from hub20.apps.ethereum_money.factories import (
    EthereumTokenValueModelFactory,
    Erc20TokenValueModelFactory,
    EthereumAccountFactory,
)
from hub20.apps.core import models
from .base import UserFactory


factory.Faker.add_provider(EthereumProvider)


class WalletFactory(factory.django.DjangoModelFactory):
    account = factory.SubFactory(EthereumAccountFactory)

    class Meta:
        model = models.Wallet


class EthereumUserBalanceEntryFactory(EthereumTokenValueModelFactory):
    user = factory.SubFactory(UserFactory)

    class Meta:
        model = models.UserBalanceEntry


class Erc20TokenUserBalanceEntryFactory(Erc20TokenValueModelFactory):
    user = factory.SubFactory(UserFactory)

    class Meta:
        model = models.UserBalanceEntry


class UserAccountFactory(factory.Factory):
    user = factory.SubFactory(UserFactory)

    class Meta:
        model = models.UserAccount


__all__ = [
    "EthereumUserBalanceEntryFactory",
    "Erc20TokenUserBalanceEntryFactory",
    "UserAccountFactory",
    "WalletFactory",
]



from django.core.exceptions import ValidationError

from ethereum.utils import check_checksum


def validate_checksumed_address(address):
    try:
        if not check_checksum(address):
            raise ValidationError(
                "%(address)s has an invalid checksum", params={"address": address},
            )
    except:
        raise ValidationError(
            "%(address)s is not a valid ethereum address", params={"address": address},
        )



import asyncio
import logging

from django.core.management.base import BaseCommand

from hub20.apps.blockchain.app_settings import START_BLOCK_NUMBER
from hub20.apps.blockchain.models import Block, make_web3

logger = logging.getLogger(__name__)


def split_block_lists(block_numbers, group_size=25):
    for n in range(0, len(block_numbers), group_size):
        yield block_numbers[n : n + group_size]


async def make_blocks_in_range(w3, start, end, speed=25):
    chain_id = int(w3.net.version)
    chain_blocks = Block.objects.filter(chain=chain_id)

    block_range = (start, end)
    recorded_block_set = set(
        chain_blocks.filter(number__range=block_range).values_list("number", flat=True)
    )
    range_set = set(range(*block_range))
    missing_blocks = list(range_set.difference(recorded_block_set))[::-1]

    counter = 0

    logger.info(f"{len(missing_blocks)} missing blocks between {start} and {end}")

    for block_list in split_block_lists(missing_blocks, group_size=speed):

        for block_number in block_list:
            counter += 1
            if (counter % speed) == 0:
                await asyncio.sleep(1)

            Block.make_all(block_number, w3)
    else:
        await asyncio.sleep(1)


async def save_new_blocks(w3):
    current_block_number = w3.eth.blockNumber
    while True:
        logger.info(f"Current block number: {current_block_number}")
        block_number = w3.eth.blockNumber
        if block_number > current_block_number:
            Block.make_all(block_number, w3)
            current_block_number = block_number
        else:
            await asyncio.sleep(5)


async def backfill(w3):
    SCAN_SIZE = 5000
    end = w3.eth.blockNumber
    while end > START_BLOCK_NUMBER:
        start = max(end - SCAN_SIZE, START_BLOCK_NUMBER)
        await make_blocks_in_range(w3, start, end)
        end = start
    logger.info(f"Backfill complete. All blocks from {end} now recorded")


class Command(BaseCommand):
    help = "Listens to new blocks and transactions on event loop and saves on DB"

    def handle(self, *args, **options):
        w3 = make_web3()
        loop = asyncio.get_event_loop()
        try:
            loop.run_until_complete(asyncio.gather(save_new_blocks(w3), backfill(w3)))
        finally:
            loop.close()



from rest_framework import serializers

from hub20.apps.blockchain.app_settings import CHAIN_ID

from . import models
from .app_settings import TRACKED_TOKENS


class TokenValueField(serializers.DecimalField):
    def __init__(self, *args, **kw):
        kw.setdefault("max_digits", 32)
        kw.setdefault("decimal_places", 18)
        super().__init__(*args, **kw)


class CurrencyRelatedField(serializers.SlugRelatedField):
    queryset = models.EthereumToken.objects.filter(chain=CHAIN_ID, address__in=TRACKED_TOKENS)

    def __init__(self, *args, **kw):
        kw.setdefault("slug_field", "address")
        super().__init__(*args, **kw)


class EthereumTokenSerializer(serializers.ModelSerializer):
    network_id = serializers.IntegerField(source="chain")

    class Meta:
        model = models.EthereumToken
        fields = ("code", "name", "address", "network_id", "decimals")
        read_only_fields = ("code", "name", "address", "network_id", "decimals")


class HyperlinkedEthereumTokenSerializer(EthereumTokenSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="ethereum_money:token-detail", lookup_field="address"
    )

    class Meta:
        model = models.EthereumToken
        fields = ("url",) + EthereumTokenSerializer.Meta.fields
        read_only_fields = ("url",) + EthereumTokenSerializer.Meta.read_only_fields



import logging

from django.db.models.signals import post_save
from django.dispatch import receiver
from eth_utils import from_wei

from hub20.apps.blockchain.models import Transaction, TransactionLog
from hub20.apps.ethereum_money import get_ethereum_account_model
from hub20.apps.ethereum_money.models import EthereumToken, EthereumTokenAmount
from hub20.apps.ethereum_money.signals import account_deposit_received

logger = logging.getLogger(__name__)
EthereumAccount = get_ethereum_account_model()


@receiver(post_save, sender=Transaction)
def on_transaction_mined_check_for_deposit(sender, **kw):
    tx = kw["instance"]
    if kw["created"]:
        accounts = EthereumAccount.objects.all()
        accounts_by_address = {account.address: account for account in accounts}

        if tx.to_address in accounts_by_address.keys():
            ETH = EthereumToken.ETH(tx.block.chain)
            eth_amount = EthereumTokenAmount(amount=from_wei(tx.value, "ether"), currency=ETH)
            account_deposit_received.send(
                sender=Transaction,
                account=accounts_by_address[tx.to_address],
                transaction=tx,
                amount=eth_amount,
            )


@receiver(post_save, sender=TransactionLog)
def on_transaction_event_check_for_token_transfer(sender, **kw):
    tx_log = kw["instance"]

    logger.info(f"Event log for {tx_log.transaction} created")
    if kw["created"]:
        tx = tx_log.transaction
        token = EthereumToken.ERC20tokens.filter(address=tx.to_address).first()

        if token is not None:
            recipient_address, transfer_amount = token._decode_token_transfer_input(tx)
            is_token_transfer = recipient_address is not None and transfer_amount is not None

            account = EthereumAccount.objects.filter(address=recipient_address).first()
            if is_token_transfer and account:
                account_deposit_received.send(
                    sender=Transaction, account=account, transaction=tx, amount=transfer_amount
                )


@receiver(account_deposit_received, sender=Transaction)
def on_account_deposit_create_balance_entry(sender, **kw):
    account = kw["account"]
    transaction = kw["transaction"]
    amount = kw["amount"]

    try:
        account.balance_entries.create(
            amount=amount.amount, currency=amount.currency, transaction=transaction
        )
    except Exception as exc:
        logger.exception(exc)


__all__ = [
    "on_transaction_mined_check_for_deposit",
    "on_transaction_event_check_for_token_transfer",
    "on_account_deposit_create_balance_entry",
]



from django.test import TestCase
from django.urls import reverse
from rest_framework.test import APIClient

from hub20.apps.core import factories


class StoreViewTestCase(TestCase):
    def setUp(self):
        self.user = factories.UserAccountFactory()
        self.client = APIClient()
        self.store = factories.StoreFactory()

    def test_anonymous_user_can_see_store(self):
        url = reverse("hub20:store-detail", kwargs={"pk": self.store.pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["id"], str(self.store.id))


class CheckoutViewTestCase(TestCase):
    def setUp(self):
        self.checkout = factories.CheckoutFactory()

    def test_delete_request_cancels_order(self):
        url = reverse("hub20:checkout-detail", kwargs={"pk": self.checkout.pk})
        response = self.client.delete(url)
        self.checkout.refresh_from_db()
        self.assertIsNone(response.data["payment_method"])


__all__ = ["StoreViewTestCase"]



import datetime
import random
from typing import Optional

from django.contrib.postgres.fields import ArrayField
from django.db import models
from model_utils.choices import Choices
from model_utils.managers import QueryManager
from model_utils.models import StatusModel
from raiden_contracts.constants import CONTRACT_TOKEN_NETWORK, CONTRACT_TOKEN_NETWORK_REGISTRY
from raiden_contracts.contract_manager import (
    ContractManager,
    contracts_precompiled_path,
    get_contracts_deployment_info,
)
from web3 import Web3
from web3.contract import Contract

from hub20.apps.blockchain.fields import EthereumAddressField
from hub20.apps.blockchain.models import Transaction
from hub20.apps.ethereum_money.models import (
    AbstractEthereumAccount,
    EthereumToken,
    EthereumTokenAmount,
    EthereumTokenAmountField,
)

CHANNEL_STATUSES = Choices("open", "settling", "settled", "unusable", "closed", "closing")


def get_token_network_registry_contract(w3: Web3):
    chain_id = int(w3.net.version)
    manager = ContractManager(contracts_precompiled_path())

    contract_data = get_contracts_deployment_info(chain_id)
    assert contract_data
    address = contract_data["contracts"][CONTRACT_TOKEN_NETWORK_REGISTRY]["address"]

    abi = manager.get_contract_abi(CONTRACT_TOKEN_NETWORK_REGISTRY)
    return w3.eth.contract(abi=abi, address=address)


class RaidenOperationError(Exception):
    pass


class TokenNetwork(models.Model):
    address = EthereumAddressField()
    token = models.OneToOneField(EthereumToken, on_delete=models.CASCADE)

    @property
    def url(self):
        return f"{self.raiden.api_root_url}/tokens/{self.address}"

    def can_reach(self, address):
        # This is a very naive assumption. One should not assume that we can
        # reach an address just because the address has an open channel.

        # However, our main purpose is only to find out if a given address is
        # being used by raiden and that we can _try_ to use for a transfer.
        open_channels = self.channels.filter(status__status=CHANNEL_STATUSES.open)
        return open_channels.filter(participant_addresses__contains=[address]).exists()

    @property
    def events(self):
        return TokenNetworkChannelEvent.objects.filter(channel__token_network=self)

    def get_contract(self, w3: Web3):
        manager = ContractManager(contracts_precompiled_path())
        abi = manager.get_contract_abi(CONTRACT_TOKEN_NETWORK)
        return w3.eth.contract(abi=abi, address=self.address)

    @classmethod
    def make(cls, token: EthereumToken, token_network_contract: Contract):
        address = token_network_contract.functions.token_to_token_networks(token.address).call()
        token_network, _ = cls.objects.get_or_create(token=token, defaults={"address": address})
        return token_network

    def __str__(self):
        return f"{self.address} - ({self.token.get_chain_display()} {self.token.code})"


class TokenNetworkChannel(models.Model):
    token_network = models.ForeignKey(
        TokenNetwork, on_delete=models.CASCADE, related_name="channels"
    )
    identifier = models.PositiveIntegerField()
    participant_addresses = ArrayField(EthereumAddressField(), size=2)
    objects = models.Manager()

    @property
    def events(self):
        return self.tokennetworkchannelevent_set.order_by(
            "transaction__block__number", "transaction__index"
        )


class TokenNetworkChannelStatus(StatusModel):
    STATUS = CHANNEL_STATUSES
    channel = models.OneToOneField(
        TokenNetworkChannel, on_delete=models.CASCADE, related_name="status"
    )

    @classmethod
    def set_status(cls, channel: TokenNetworkChannel):
        last_event = channel.events.last()
        event_name = last_event and last_event.name
        status = event_name and {
            "ChannelOpened": CHANNEL_STATUSES.open,
            "ChannelClosed": CHANNEL_STATUSES.closed,
        }.get(event_name)
        cls.objects.update_or_create(channel=channel, defaults={"status": status})


class TokenNetworkChannelEvent(models.Model):
    channel = models.ForeignKey(TokenNetworkChannel, on_delete=models.CASCADE)
    transaction = models.ForeignKey(Transaction, on_delete=models.CASCADE)
    name = models.CharField(max_length=32, db_index=True)

    class Meta:
        unique_together = ("channel", "transaction")


class Raiden(AbstractEthereumAccount):
    url = models.URLField(help_text="Root URL of server (without api/v1)")
    token_networks = models.ManyToManyField(TokenNetwork, blank=True)

    @property
    def api_root_url(self):
        url = self.url.strip("/")
        return f"{url}/api/v1"

    def __str__(self):
        return f"Raiden @ {self.address}"


class Channel(StatusModel):
    STATUS = CHANNEL_STATUSES
    raiden = models.ForeignKey(Raiden, on_delete=models.CASCADE, related_name="channels")
    token_network = models.ForeignKey(TokenNetwork, on_delete=models.CASCADE)
    identifier = models.PositiveIntegerField()
    partner_address = EthereumAddressField(db_index=True)
    balance = EthereumTokenAmountField()
    total_deposit = EthereumTokenAmountField()
    total_withdraw = EthereumTokenAmountField()

    objects = models.Manager()
    funded = QueryManager(status=STATUS.open, balance__gt=0)
    available = QueryManager(status=STATUS.open)

    @property
    def url(self):
        return f"{self.raiden.api_root_url}/channels/{self.token.address}/{self.partner_address}"

    @property
    def payments_url(self):
        return f"{self.raiden.api_root_url}/payments/{self.token.address}/{self.partner_address}"

    @property
    def token(self):
        return self.token_network.token

    @property
    def balance_amount(self) -> EthereumTokenAmount:
        return EthereumTokenAmount(amount=self.balance, currency=self.token)

    @property
    def last_event_timestamp(self) -> Optional[datetime.datetime]:
        latest_event = self.payments.order_by("-timestamp").first()
        return latest_event and latest_event.timestamp

    def send(self, address, transfer_amount: EthereumTokenAmount):
        raise NotImplementedError("TODO")

    def __str__(self):
        return f"Channel {self.identifier} ({self.partner_address})"

    @classmethod
    def make(cls, raiden: Raiden, **channel_data):
        token_network = TokenNetwork.objects.get(address=channel_data.pop("token_network_address"))
        token = token_network.token

        assert token.address is not None
        assert token.address == channel_data.pop("token_address")

        balance = token.from_wei(channel_data.pop("balance"))
        total_deposit = token.from_wei(channel_data.pop("total_deposit"))
        total_withdraw = token.from_wei(channel_data.pop("total_withdraw"))

        channel, _ = raiden.channels.update_or_create(
            identifier=channel_data["channel_identifier"],
            partner_address=channel_data["partner_address"],
            token_network=token_network,
            defaults={
                "status": channel_data["state"],
                "balance": balance.amount,
                "total_deposit": total_deposit.amount,
                "total_withdraw": total_withdraw.amount,
            },
        )
        return channel

    @classmethod
    def select_for_transfer(cls, recipient_address, transfer_amount: EthereumTokenAmount):
        funded = cls.objects.filter(
            token_network__token=transfer_amount.currency, balance__gte=transfer_amount.amount
        )
        reachable = [c for c in funded if c.token_network.can_reach(recipient_address)]
        return random.choice(reachable) if reachable else None

    class Meta:
        unique_together = (
            ("raiden", "token_network", "partner_address"),
            ("raiden", "token_network", "identifier"),
        )


class Payment(models.Model):
    channel = models.ForeignKey(Channel, on_delete=models.CASCADE, related_name="payments")
    amount = EthereumTokenAmountField()
    timestamp = models.DateTimeField()
    identifier = models.BigIntegerField()
    sender_address = EthereumAddressField()
    receiver_address = EthereumAddressField()

    @property
    def url(self):
        return f"{self.raiden.api_root_url}/{self.token.address}/{self.partner_address}"

    @property
    def token(self):
        return self.channel.token

    @property
    def as_token_amount(self) -> EthereumTokenAmount:
        return EthereumTokenAmount(amount=self.amount, currency=self.token)

    @classmethod
    def make(cls, channel: Channel, **payment_data):
        payment, _ = channel.payments.get_or_create(channel=channel, **payment_data)
        return payment

    class Meta:
        unique_together = ("channel", "timestamp", "sender_address", "receiver_address")


__all__ = ["TokenNetwork", "Raiden", "TokenNetwork", "Channel", "Payment"]



from django.conf import settings

CHAIN_ID = int(getattr(settings, "BLOCKCHAIN_NETWORK_ID", None) or (5 if settings.DEBUG else 1))
START_BLOCK_NUMBER = int(getattr(settings, "BLOCKCHAIN_START_BLOCK_NUMBER", 0) or 0)



import logging

from django.db.models.signals import post_save
from django.dispatch import receiver

from hub20.apps.raiden.models import Payment, TokenNetworkChannelStatus, TokenNetworkChannelEvent
from hub20.apps.raiden.signals import raiden_payment_received


logger = logging.getLogger(__name__)


@receiver(post_save, sender=Payment)
def on_payment_created_check_received(sender, **kw):
    payment = kw["instance"]
    if kw["created"]:
        if payment.receiver_address == payment.channel.raiden.address:
            logger.info(f"New payment received by {payment.channel}")
            raiden_payment_received.send(sender=Payment, payment=payment)


@receiver(post_save, sender=TokenNetworkChannelEvent)
def on_token_network_channel_event_set_status(sender, **kw):
    event = kw["instance"]
    if kw["created"]:
        TokenNetworkChannelStatus.set_status(event.channel)



# Generated by Django 2.2.9 on 2020-01-30 13:14

from django.db import migrations, models
import django.db.models.deletion
import hub20.apps.blockchain.fields
import hub20.apps.ethereum_money.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('blockchain', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='EthereumAccount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('address', hub20.apps.blockchain.fields.EthereumAddressField(db_index=True, unique=True)),
                ('private_key', hub20.apps.blockchain.fields.HexField(max_length=64, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='EthereumToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('chain', models.PositiveIntegerField(choices=[(1, 'Mainnet'), (2, 'Test Network'), (3, 'Ropsten'), (4, 'Rinkeby'), (5, 'Görli'), (42, 'Kovan')])),
                ('code', models.CharField(max_length=8)),
                ('name', models.CharField(max_length=500)),
                ('decimals', models.PositiveIntegerField(default=18)),
                ('address', hub20.apps.blockchain.fields.EthereumAddressField(default='0x0000000000000000000000000000000000000000')),
            ],
            options={
                'unique_together': {('chain', 'address')},
            },
        ),
        migrations.CreateModel(
            name='AccountBalanceEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', hub20.apps.ethereum_money.models.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='balance_entries', to='ethereum_money.EthereumAccount')),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ethereum_money.EthereumToken')),
                ('transaction', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='blockchain.Transaction')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import factory
import factory.fuzzy
from django.utils import timezone

from ..models import Block, Transaction, TransactionLog
from .providers import EthereumProvider

factory.Faker.add_provider(EthereumProvider)


TEST_CHAIN_ID = 2


def find_parent_by_block_number(block):
    return Block.objects.filter(chain=block.chain, number=block.number - 1).first()


def make_parent_hash(block):
    parent = find_parent_by_block_number(block)
    return parent and parent.hash or factory.Faker("hex64").generate()


class BlockFactory(factory.django.DjangoModelFactory):
    chain = TEST_CHAIN_ID
    hash = factory.Faker("hex64")
    number = factory.Sequence(lambda n: n)
    timestamp = factory.LazyAttribute(lambda obj: timezone.now())
    parent_hash = factory.LazyAttribute(lambda obj: make_parent_hash(obj))
    uncle_hashes = []

    class Meta:
        model = Block


class TransactionFactory(factory.django.DjangoModelFactory):
    block = factory.SubFactory(BlockFactory)
    hash = factory.Faker("hex64")
    from_address = factory.Faker("ethereum_address")
    to_address = factory.Faker("ethereum_address")
    gas = 21000
    gas_price = factory.fuzzy.FuzzyInteger(1e10, 2e12)
    nonce = factory.Sequence(lambda n: n)
    index = factory.Faker("uint256")
    value = factory.fuzzy.FuzzyInteger(1e15, 1e21)

    class Meta:
        model = Transaction


class TransactionLogFactory(factory.django.DjangoModelFactory):
    transaction = factory.SubFactory(TransactionFactory)
    index = factory.Sequence(lambda n: n)
    topics = ["0x0", "0x0", "0x0"]
    data = "0x0"

    class Meta:
        model = TransactionLog



import logging

from asgiref.sync import async_to_sync
from celery import shared_task
from channels.layers import get_channel_layer
from django.db import transaction
from django.utils import timezone

from .consumers import CheckoutConsumer
from .models import PaymentOrderEvent, PaymentOrderMethod, Transfer

logger = logging.getLogger(__name__)


@shared_task
def execute_transfer(transfer_id):
    transfer = Transfer.objects.get_subclass(id=transfer_id)
    transfer.execute()


@shared_task
def expire_payment_method(payment_method_id):

    payment_method = PaymentOrderMethod.objects.filter(id=payment_method_id).first()

    if not payment_method:
        return

    if payment_method.expiration_time >= timezone.now():
        return

    with transaction.atomic():
        order = payment_method.order
        if not order.is_finalized:
            order.events.create(status=PaymentOrderEvent.STATUS.expired)
        payment_method.delete()


@shared_task
def publish_checkout_event(checkout_id, event="checkout.event", **event_data):
    layer = get_channel_layer()
    channel_group_name = CheckoutConsumer.get_group_name(checkout_id)

    logger.info(f"Publishing event {event}. Data: {event_data}")

    event_data.update({"type": "publish_payment_event", "event_name": event})

    async_to_sync(layer.group_send)(channel_group_name, event_data)



from typing import Callable

from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.urls import include, path
from rest_auth.views import (
    LoginView,
    LogoutView,
    UserDetailsView,
    PasswordChangeView,
    PasswordResetView,
    PasswordResetConfirmView,
)


def make_auth_view(url_path: str, view_class, view_name: str) -> Callable:
    return path(url_path, view_class.as_view(), name=view_name)


urlpatterns = [
    # URLs that do not require a session or valid token
    make_auth_view("api/accounts/password/reset", PasswordResetView, "rest_password_reset"),
    make_auth_view(
        "api/accounts/password/confirm", PasswordResetConfirmView, "rest_password_reset_confirm",
    ),
    make_auth_view("api/accounts/password/change", PasswordChangeView, "rest_password_change"),
    make_auth_view("api/accounts/user", UserDetailsView, "rest_user_details"),
    make_auth_view("api/session/login", LoginView, "rest_login"),
    make_auth_view("api/session/logout", LogoutView, "rest_logout"),
    path("api/register/", include("rest_auth.registration.urls")),
    path("api/tokens/", include("hub20.apps.ethereum_money.api", namespace="ethereum_money")),
    path("api/", include("hub20.apps.core.api", namespace="hub20")),
    path("admin/", admin.site.urls),
]

urlpatterns.extend(static(settings.STATIC_URL, document_root=settings.STATIC_ROOT))



# Generated by Django 2.2.9 on 2020-01-30 14:22

from django.conf import settings
import django.contrib.sites.models
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import hub20.apps.blockchain.fields
import hub20.apps.core.models.payments
import hub20.apps.ethereum_money.models
import model_utils.fields
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('blockchain', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('raiden', '0001_initial'),
        ('ethereum_money', '0001_initial'),
        ('sites', '0002_alter_domain_unique'),
    ]

    operations = [
        migrations.CreateModel(
            name='PaymentOrder',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', hub20.apps.ethereum_money.models.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ethereum_money.EthereumToken')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Store',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=300)),
                ('url', models.URLField()),
                ('accepted_currencies', models.ManyToManyField(to='ethereum_money.EthereumToken')),
                ('owner', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Transfer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', hub20.apps.ethereum_money.models.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('memo', models.TextField(blank=True, null=True)),
                ('identifier', models.CharField(blank=True, max_length=300, null=True)),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ethereum_money.EthereumToken')),
                ('sender', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='transfers_sent', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UserReserve',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', hub20.apps.ethereum_money.models.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ethereum_money.EthereumToken')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='reserves', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='HubSite',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('sites.site',),
            managers=[
                ('objects', django.contrib.sites.models.SiteManager()),
            ],
        ),
        migrations.CreateModel(
            name='ExternalTransfer',
            fields=[
                ('transfer_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.Transfer')),
                ('recipient_address', hub20.apps.blockchain.fields.EthereumAddressField(db_index=True)),
            ],
            options={
                'abstract': False,
            },
            bases=('core.transfer',),
        ),
        migrations.CreateModel(
            name='Wallet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('account', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='ethereum_money.EthereumAccount')),
            ],
        ),
        migrations.CreateModel(
            name='UserBalanceEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', hub20.apps.ethereum_money.models.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ethereum_money.EthereumToken')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='balance_entries', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='StoreRSAKeyPair',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('public_key_pem', models.TextField()),
                ('private_key_pem', models.TextField()),
                ('store', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='rsa', to='core.Store')),
            ],
        ),
        migrations.CreateModel(
            name='PaymentOrderMethod',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('expiration_time', models.DateTimeField()),
                ('identifier', models.BigIntegerField(default=hub20.apps.core.models.payments.generate_payment_order_id, unique=True)),
                ('order', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='payment_method', to='core.PaymentOrder')),
                ('raiden', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='raiden.Raiden')),
                ('wallet', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='core.Wallet')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Payment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('amount', hub20.apps.ethereum_money.models.EthereumTokenAmountField(decimal_places=18, max_digits=32)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('currency', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='ethereum_money.EthereumToken')),
                ('order', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='core.PaymentOrder')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Checkout',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('external_identifier', models.TextField()),
                ('requester_ip', models.GenericIPAddressField(null=True)),
                ('payment_order', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.PaymentOrder')),
                ('store', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.Store')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BlockchainTransaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('transaction_hash', hub20.apps.blockchain.fields.HexField(db_index=True, max_length=64, unique=True)),
                ('transfer', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='chain_transaction', to='core.Transfer')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UserTransferReserve',
            fields=[
                ('userreserve_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.UserReserve')),
                ('transfer', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='reserve', to='core.Transfer')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.userreserve',),
        ),
        migrations.CreateModel(
            name='TransferEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('status', model_utils.fields.StatusField(choices=[('scheduled', 'scheduled'), ('failed', 'failed'), ('canceled', 'canceled'), ('executed', 'executed'), ('confirmed', 'confirmed')], default='scheduled', max_length=100, no_check_for_status=True, verbose_name='status')),
                ('status_changed', model_utils.fields.MonitorField(default=django.utils.timezone.now, monitor='status', verbose_name='status changed')),
                ('transfer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='events', to='core.Transfer')),
            ],
            options={
                'get_latest_by': 'created',
                'unique_together': {('transfer', 'status')},
            },
        ),
        migrations.CreateModel(
            name='RaidenTransaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('identifier', models.PositiveIntegerField()),
                ('channel', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='raiden.Channel')),
                ('transfer', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='raiden_transaction', to='core.Transfer')),
            ],
            options={
                'unique_together': {('channel', 'identifier')},
            },
        ),
        migrations.CreateModel(
            name='RaidenPayment',
            fields=[
                ('payment_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.Payment')),
                ('payment', models.OneToOneField(on_delete=django.db.models.deletion.PROTECT, to='raiden.Payment')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.payment',),
        ),
        migrations.CreateModel(
            name='PaymentOrderEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('status', model_utils.fields.StatusField(choices=[('requested', 'requested'), ('partial', 'partial'), ('received', 'received'), ('confirmed', 'confirmed'), ('expired', 'expired'), ('canceled', 'canceled'), ('voided', 'voided')], default='requested', max_length=100, no_check_for_status=True, verbose_name='status')),
                ('status_changed', model_utils.fields.MonitorField(default=django.utils.timezone.now, monitor='status', verbose_name='status changed')),
                ('order', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='events', to='core.PaymentOrder')),
            ],
            options={
                'get_latest_by': 'created',
                'unique_together': {('order', 'status')},
            },
        ),
        migrations.CreateModel(
            name='InternalTransfer',
            fields=[
                ('transfer_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.Transfer')),
                ('receiver', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='transfers_received', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
            bases=('core.transfer',),
        ),
        migrations.CreateModel(
            name='InternalPayment',
            fields=[
                ('payment_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.Payment')),
                ('memo', models.TextField(blank=True, null=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
            bases=('core.payment',),
        ),
        migrations.CreateModel(
            name='BlockchainPayment',
            fields=[
                ('payment_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.Payment')),
                ('transaction', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='blockchain.Transaction')),
            ],
            options={
                'abstract': False,
            },
            bases=('core.payment',),
        ),
    ]



from django.dispatch import Signal

blockchain_payment_sent = Signal(providing_args=["recipient", "amount", "transaction_hash"])
payment_received = Signal(providing_args=["payment"])
payment_confirmed = Signal(providing_args=["payment"])
order_canceled = Signal(providing_args=["payment_order"])
order_paid = Signal(providing_args=["payment_order"])
transfer_scheduled = Signal(providing_args=["transfer"])
transfer_confirmed = Signal(providing_args=["transfer"])
transfer_executed = Signal(providing_args=["transfer"])
transfer_failed = Signal(providing_args=["transfer", "reason"])

__all__ = [
    "blockchain_payment_sent",
    "payment_received",
    "payment_confirmed",
    "order_canceled",
    "order_paid",
    "transfer_scheduled",
    "transfer_confirmed",
    "transfer_executed",
    "transfer_failed",
]



import logging

from django.core.management.base import BaseCommand
from eth_utils import to_checksum_address

from hub20.apps.ethereum_money.app_settings import TRACKED_TOKENS
from hub20.apps.ethereum_money.models import EthereumToken

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Loads data relevant to all tokens that are going to be used by the instance"

    def handle(self, *args, **options):
        for token_address in TRACKED_TOKENS:
            logger.info(f"Checking token {token_address}...")
            try:
                EthereumToken.make(to_checksum_address(token_address))
            except OverflowError:
                logger.error(f"{token_address} is not a valid address or not ERC20-compliant")
            except Exception as exc:
                logger.exception(f"Failed to load token data for {token_address}", exc_info=exc)



import random
from datetime import timedelta

from django.utils import timezone
import factory
from factory import fuzzy

from hub20.apps.blockchain.factories import EthereumProvider
from hub20.apps.ethereum_money.factories import Erc20TokenFactory

from . import models


factory.Faker.add_provider(EthereumProvider)


class TokenNetworkFactory(factory.django.DjangoModelFactory):
    address = factory.Faker("ethereum_address")
    token = factory.SubFactory(Erc20TokenFactory)

    class Meta:
        model = models.TokenNetwork


class RaidenFactory(factory.django.DjangoModelFactory):
    url = factory.Sequence(lambda n: f"http://raiden{n:02}.example.org")
    address = factory.Faker("ethereum_address")

    @factory.post_generation
    def token_networks(self, create, extracted, **kw):
        if not create:
            return

        token_networks = extracted or []
        for network in token_networks:
            self.token_networks.add(network)

    class Meta:
        model = models.Raiden


class ChannelFactory(factory.django.DjangoModelFactory):
    raiden = factory.SubFactory(RaidenFactory)
    token_network = factory.SubFactory(TokenNetworkFactory)
    partner_address = factory.Faker("ethereum_address")
    identifier = factory.Sequence(lambda n: n)
    balance = fuzzy.FuzzyDecimal(1, 100, precision=6)
    total_deposit = factory.LazyAttribute(lambda obj: obj.balance)
    total_withdraw = 0
    status = "open"

    class Meta:
        model = models.Channel


class PaymentEventFactory(factory.django.DjangoModelFactory):
    channel = factory.SubFactory(ChannelFactory)
    amount = factory.fuzzy.FuzzyDecimal(0, 1, precision=4)
    timestamp = factory.LazyAttribute(
        lambda obj: PaymentEventFactory.make_sequenced_timestamp(obj)
    )
    identifier = factory.Sequence(lambda n: n)
    sender_address = factory.Faker("ethereum_address")
    receiver_address = factory.Faker("ethereum_address")

    @staticmethod
    def make_sequenced_timestamp(payment_event):
        since = payment_event.channel.last_event_timestamp or timezone.now()
        return since + timedelta(microseconds=random.randint(1, 1e8))

    class Meta:
        model = models.Payment


__all__ = ["TokenNetworkFactory", "RaidenFactory", "ChannelFactory", "PaymentEventFactory"]



# Generated by Django 2.2.9 on 2020-01-28 12:02

import django.contrib.postgres.fields
from django.db import migrations, models
import django.db.models.deletion
import hub20.apps.blockchain.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Block',
            fields=[
                ('hash', hub20.apps.blockchain.fields.HexField(max_length=64, primary_key=True, serialize=False)),
                ('chain', models.PositiveIntegerField(choices=[(1, 'Mainnet'), (2, 'Test Network'), (3, 'Ropsten'), (4, 'Rinkeby'), (5, 'Görli'), (42, 'Kovan')], default=1)),
                ('number', models.PositiveIntegerField(db_index=True)),
                ('timestamp', models.DateTimeField()),
                ('parent_hash', hub20.apps.blockchain.fields.HexField(max_length=64)),
                ('uncle_hashes', django.contrib.postgres.fields.ArrayField(base_field=hub20.apps.blockchain.fields.HexField(max_length=64), size=None)),
            ],
            options={
                'unique_together': {('chain', 'hash', 'number')},
            },
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('hash', hub20.apps.blockchain.fields.HexField(db_index=True, max_length=64)),
                ('from_address', hub20.apps.blockchain.fields.EthereumAddressField(db_index=True)),
                ('to_address', hub20.apps.blockchain.fields.EthereumAddressField(db_index=True, null=True)),
                ('gas', hub20.apps.blockchain.fields.Uint256Field()),
                ('gas_price', hub20.apps.blockchain.fields.Uint256Field()),
                ('nonce', hub20.apps.blockchain.fields.Uint256Field()),
                ('index', hub20.apps.blockchain.fields.Uint256Field()),
                ('value', hub20.apps.blockchain.fields.Uint256Field()),
                ('data', models.TextField()),
                ('block', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='blockchain.Block')),
            ],
        ),
        migrations.CreateModel(
            name='TransactionLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('index', models.SmallIntegerField()),
                ('data', models.TextField()),
                ('topics', django.contrib.postgres.fields.ArrayField(base_field=models.TextField(), size=None)),
                ('transaction', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='logs', to='blockchain.Transaction')),
            ],
            options={
                'unique_together': {('transaction', 'index')},
            },
        ),
    ]



import datetime
import uuid
from enum import Enum
from typing import Optional

import jwt
from Crypto.PublicKey import RSA
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from model_utils.models import TimeStampedModel

from hub20.apps.ethereum_money.models import EthereumToken

from .payments import PaymentOrder


class CheckoutEvents(Enum):
    BLOCKCHAIN_BLOCK_CREATED = "blockchain.block.created"
    BLOCKCHAIN_TRANSFER_BROADCAST = "blockchain.transfer.broadcast"
    BLOCKCHAIN_TRANSFER_MINED = "blockchain.transfer.mined"
    RAIDEN_TRANSFER_RECEIVED = "raiden.transfer.received"


class Store(models.Model):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    owner = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    name = models.CharField(max_length=300)
    url = models.URLField()
    accepted_currencies = models.ManyToManyField(EthereumToken)

    def __str__(self):
        return f"{self.name} ({self.url})"


class StoreRSAKeyPair(models.Model):
    DEFAULT_KEY_SIZE = 2048

    store = models.OneToOneField(Store, on_delete=models.CASCADE, related_name="rsa")
    public_key_pem = models.TextField()
    private_key_pem = models.TextField()

    @classmethod
    def generate(cls, store: Store, keysize: Optional[int] = None):
        bits = keysize or cls.DEFAULT_KEY_SIZE
        key = RSA.generate(bits)
        public_key_pem = key.publickey().export_key().decode()
        private_key_pem = key.export_key().decode()

        pair, _ = cls.objects.update_or_create(
            store=store,
            defaults={"public_key_pem": public_key_pem, "private_key_pem": private_key_pem},
        )
        return pair


class Checkout(TimeStampedModel):
    id = models.UUIDField(default=uuid.uuid4, primary_key=True)
    store = models.ForeignKey(Store, on_delete=models.CASCADE)
    payment_order = models.ForeignKey(PaymentOrder, on_delete=models.CASCADE)
    external_identifier = models.TextField()
    requester_ip = models.GenericIPAddressField(null=True)

    def clean(self):
        if self.store.owner != self.payment_order.user:
            raise ValidationError("Creator or payment order must be the same as store owner")

        if self.payment_order.currency not in self.store.accepted_currencies.all():
            raise ValidationError(
                f"{self.store.name} does not accept payment in {self.payment_order.currency}"
            )

    def issue_voucher(self, **data):
        data.update(
            {
                "iat": datetime.datetime.utcnow(),
                "iss": self.external_identifier,
                "status": self.payment_order.status,
                "token": {
                    "symbol": self.payment_order.currency.code,
                    "address": self.payment_order.currency.address,
                },
                "payments": [
                    {
                        "amount": str(p.amount),
                        "confirmed": p.is_confirmed,
                        "identifier": p.identifier,
                        "route": p.route,
                    }
                    for p in self.payment_order.payments
                ],
                "total_amount": str(self.payment_order.amount),
                "total_confirmed": str(self.payment_order.total_confirmed),
                "payment_order_id": self.payment_order.id,
                "checkout_id": str(self.id),
            }
        )

        private_key = self.store.rsa.private_key_pem.encode()
        return jwt.encode(data, private_key, algorithm="RS256").decode()


__all__ = ["Store", "StoreRSAKeyPair", "Checkout"]



from channels.routing import ProtocolTypeRouter, URLRouter
from django import setup

setup()

from hub20.apps.core.api import consumer_patterns  # isort:skip
from .middleware import TokenAuthMiddlewareStack  # isort:skip


application = ProtocolTypeRouter(
    {"websocket": TokenAuthMiddlewareStack(URLRouter(consumer_patterns))}
)



from model_utils import Choices

PAYMENT_EVENT_TYPES = Choices(
    "requested", "partial", "received", "confirmed", "expired", "canceled", "voided"
)
TRANSFER_EVENT_TYPES = Choices("scheduled", "failed", "canceled", "executed", "confirmed")
PAYMENT_METHODS = Choices("blockchain", "raiden", "internal")



from __future__ import annotations

import logging
import os
from decimal import Decimal
from typing import Any, Tuple, Union

import ethereum
from django.conf import settings
from django.db import models
from django.db.models import Q, Sum
from eth_utils import to_checksum_address
from ethereum.abi import ContractTranslator
from ethtoken import token
from ethtoken.abi import EIP20_ABI
from model_utils.managers import QueryManager
from web3 import Web3

from hub20.apps.blockchain.app_settings import CHAIN_ID
from hub20.apps.blockchain.choices import ETHEREUM_CHAINS
from hub20.apps.blockchain.fields import EthereumAddressField, HexField
from hub20.apps.blockchain.models import Transaction, make_web3
from hub20.apps.ethereum_money.app_settings import TRANSFER_GAS_LIMIT

logger = logging.getLogger(__name__)


def get_max_fee() -> EthereumTokenAmount:
    w3 = make_web3()
    ETH = EthereumToken.ETH(chain_id=int(w3.net.version))

    gas_price = w3.eth.generateGasPrice()
    return ETH.from_wei(TRANSFER_GAS_LIMIT * gas_price)


def encode_transfer_data(recipient_address, amount: EthereumTokenAmount):
    translator = ContractTranslator(EIP20_ABI)
    encoded_data = translator.encode_function_call("transfer", (recipient_address, amount.as_wei))
    return f"0x{encoded_data.hex()}"


class AbstractEthereumAccount(models.Model):
    address = EthereumAddressField(unique=True, db_index=True)

    def send(self, recipient_address, transfer_amount: EthereumTokenAmount) -> str:
        raise NotImplementedError()

    def sign_transaction(self, transaction_data, **kw):
        raise NotImplementedError()

    class Meta:
        abstract = True


class EthereumAccount(AbstractEthereumAccount):
    private_key = HexField(max_length=64, unique=True)

    def send(self, recipient_address, transfer_amount: EthereumTokenAmount, *args, **kw) -> str:
        w3 = make_web3()
        transaction_data = transfer_amount.currency.build_transfer_transaction(
            w3=w3, sender=self.address, recipient=recipient_address, amount=transfer_amount
        )
        signed_tx = self.sign_transaction(transaction_data=transaction_data, w3=w3)
        return w3.eth.sendRawTransaction(signed_tx.rawTransaction)

    def sign_transaction(self, transaction_data, *args, **kw):
        w3 = kw.get("w3") or make_web3()
        return w3.eth.account.signTransaction(transaction_data, self.private_key)

    @classmethod
    def generate(cls):
        private_key = os.urandom(32)
        address = ethereum.utils.privtoaddr(private_key)
        checksum_address = ethereum.utils.checksum_encode(address.hex())
        return cls.objects.create(address=checksum_address, private_key=private_key.hex())


class EthereumToken(models.Model):
    NULL_ADDRESS = "0x0000000000000000000000000000000000000000"
    chain = models.PositiveIntegerField(choices=ETHEREUM_CHAINS)
    code = models.CharField(max_length=8)
    name = models.CharField(max_length=500)
    decimals = models.PositiveIntegerField(default=18)
    address = EthereumAddressField(default=NULL_ADDRESS)

    objects = models.Manager()
    ERC20tokens = QueryManager(~Q(address=NULL_ADDRESS))
    ethereum = QueryManager(address=NULL_ADDRESS)

    @property
    def is_ERC20(self) -> bool:
        return self.address != self.NULL_ADDRESS

    def __str__(self) -> str:
        components = [self.code]
        if self.is_ERC20:
            components.append(self.address)

        components.append(self.get_chain_display())
        return " - ".join(components)

    def build_transfer_transaction(self, w3: Web3, sender, recipient, amount: EthereumTokenAmount):

        chain_id = int(w3.net.version)
        message = f"Web3 client is on network {chain_id}, token {self.code} is on {self.chain}"
        assert self.chain == chain_id, message

        transaction_params = {
            "chainId": chain_id,
            "nonce": w3.eth.getTransactionCount(sender),
            "gasPrice": w3.eth.generateGasPrice(),
            "gas": TRANSFER_GAS_LIMIT,
            "from": sender,
        }

        if self.is_ERC20:
            transaction_params.update(
                {"to": self.address, "value": 0, "data": encode_transfer_data(recipient, amount)}
            )
        else:
            transaction_params.update({"to": recipient, "value": amount.as_wei})
        return transaction_params

    def _decode_token_transfer_input(self, transaction: Transaction) -> Tuple:
        # A transfer transaction input is 'function,address,uint256'
        # i.e, 16 bytes + 20 bytes + 32 bytes = hex string of length 136
        try:
            # transaction input strings are '0x', so we they should be 138 chars long
            assert len(transaction.data) == 138, "Not a ERC20 transfer transaction"
            assert transaction.logs.count() == 1, "Transaction does not contain log changes"

            recipient_address = to_checksum_address(transaction.data[-104:-64])

            wei_transferred = int(transaction.data[-64:], 16)
            tx_log = transaction.logs.first()

            assert int(tx_log.data, 16) == wei_transferred, "Log data and tx amount do not match"

            return recipient_address, self.from_wei(wei_transferred)
        except AssertionError as exc:
            logger.info(f"Failed to get transfer data from transaction: {exc}")
            return None, None
        except ValueError:
            logger.info(f"Failed to extract transfer amounts from {transaction.hash.hex()}")
            return None, None
        except Exception as exc:
            logger.exception(exc)
            return None, None

    def from_wei(self, wei_amount: int) -> EthereumTokenAmount:
        value = Decimal(wei_amount) / (10 ** self.decimals)
        return EthereumTokenAmount(amount=value, currency=self)

    @staticmethod
    def ETH(chain_id: int):
        eth, _ = EthereumToken.objects.get_or_create(
            chain=chain_id, code="ETH", defaults={"name": "Ethereum"}
        )
        return eth

    @classmethod
    def make(cls, address: str, chain_id: int = CHAIN_ID):
        if chain_id != CHAIN_ID:
            raise ValueError(
                f"Can not make token on chain {chain_id} while connected to {CHAIN_ID}"
            )

        if address == EthereumToken.NULL_ADDRESS:
            return EthereumToken.ETH(chain_id)

        proxy = token(address)

        obj, _ = cls.objects.update_or_create(
            address=address,
            chain=chain_id,
            defaults={"name": proxy.name(), "code": proxy.symbol(), "decimals": proxy.decimals()},
        )
        return obj

    class Meta:
        unique_together = (("chain", "address"),)


class EthereumTokenAmountField(models.DecimalField):
    def __init__(self, *args: Any, **kw: Any) -> None:
        kw.setdefault("decimal_places", 18)
        kw.setdefault("max_digits", 32)

        super().__init__(*args, **kw)


class EthereumTokenValueModel(models.Model):
    amount = EthereumTokenAmountField()
    currency = models.ForeignKey(EthereumToken, on_delete=models.PROTECT)

    @property
    def as_token_amount(self):
        return EthereumTokenAmount(amount=self.amount, currency=self.currency)

    @property
    def formatted_amount(self):
        return self.as_token_amount.formatted

    class Meta:
        abstract = True


class AccountBalanceEntry(EthereumTokenValueModel):
    account = models.ForeignKey(
        settings.ETHEREUM_ACCOUNT_MODEL, on_delete=models.CASCADE, related_name="balance_entries"
    )
    transaction = models.OneToOneField(Transaction, on_delete=models.CASCADE)


class EthereumTokenAmount:
    def __init__(self, amount: Union[int, str, Decimal], currency: EthereumToken):
        self.amount: Decimal = Decimal(amount)
        self.currency: EthereumToken = currency

    @property
    def formatted(self):
        return f"{self.amount} {self.currency.code}"

    @property
    def as_wei(self) -> int:
        return int(self.amount * (10 ** self.currency.decimals))

    @property
    def as_hex(self) -> str:
        return hex(self.as_wei)

    @property
    def is_ETH(self) -> bool:
        return self.currency.address == EthereumToken.NULL_ADDRESS

    def _check_currency_type(self, other: EthereumTokenAmount):
        if not self.currency == other.currency:
            raise ValueError(f"Can not operate {self.currency} and {other.currency}")

    def __add__(self, other: EthereumTokenAmount) -> EthereumTokenAmount:
        self._check_currency_type(self)
        return self.__class__(self.amount + other.amount, self.currency)

    def __mul__(self, other: Union[int, Decimal]) -> EthereumTokenAmount:
        return EthereumTokenAmount(amount=Decimal(other * self.amount), currency=self.currency)

    def __rmul__(self, other: Union[int, Decimal]) -> EthereumTokenAmount:
        return self.__mul__(other)

    def __eq__(self, other: object) -> bool:
        message = f"Can not compare {self.currency} amount with {type(other)}"
        assert isinstance(other, EthereumTokenAmount), message

        return self.currency == other.currency and self.amount == other.amount

    def __lt__(self, other: EthereumTokenAmount):
        self._check_currency_type(other)
        return self.amount < other.amount

    def __le__(self, other: EthereumTokenAmount):
        self._check_currency_type(other)
        return self.amount <= other.amount

    def __gt__(self, other: EthereumTokenAmount):
        self._check_currency_type(other)
        return self.amount > other.amount

    def __ge__(self, other: EthereumTokenAmount):
        self._check_currency_type(other)
        return self.amount >= other.amount

    def __str__(self):
        return self.formatted

    def __repr__(self):
        return self.formatted

    @classmethod
    def aggregated(cls, queryset, currency: EthereumToken):
        entries = queryset.filter(currency=currency)
        amount = entries.aggregate(total=Sum("amount")).get("total") or Decimal(0)
        return cls(amount=amount, currency=currency)


__all__ = [
    "EthereumToken",
    "EthereumTokenAmount",
    "EthereumTokenValueModel",
    "EthereumAccount",
    "AccountBalanceEntry",
    "get_max_fee",
    "encode_transfer_data",
]
